# -*- coding: utf-8 -*-
"""
It is better to use constant variables instead of hoping you spell the same
string correctly every time you use it. (Also it makes it much easier if a
string name changes)
"""
import logging
import six
import numpy as np
import math
import utool as ut
from collections import OrderedDict
from os.path import join

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


CONTAINERIZED = ut.get_argflag('--containerized')
PRODUCTION = ut.get_argflag('--production')
HTTPS = ut.get_argflag('--https')


CONTAINER_NAME = ut.get_argval(
    '--container-name', type_=str, default=ut.get_computer_name()
)
ENGINE_SLOT = ut.get_argval('--engine-slot', type_=str, default='default')


PI = math.pi
TAU = 2.0 * PI

VIEWTEXT_TO_YAW_RADIANS = OrderedDict(
    [
        ('right', 0.000 * TAU),
        ('frontright', 0.125 * TAU),
        ('front', 0.250 * TAU),
        ('frontleft', 0.375 * TAU),
        ('left', 0.500 * TAU),
        ('backleft', 0.625 * TAU),
        ('back', 0.750 * TAU),
        ('backright', 0.875 * TAU),
    ]
)

# Mapping of viewpoints codes to yaw angles
VIEWTEXT_TO_YAW_RADIANS = OrderedDict(
    [
        ('right', 0.000 * TAU),
        ('frontright', 0.125 * TAU),
        ('front', 0.250 * TAU),
        ('frontleft', 0.375 * TAU),
        ('left', 0.500 * TAU),
        ('backleft', 0.625 * TAU),
        ('back', 0.750 * TAU),
        ('backright', 0.875 * TAU),
    ]
)
VIEWTEXT_TO_VIEWPOINT_RADIANS = VIEWTEXT_TO_YAW_RADIANS


# TODO: DEPRICATE ALL THE YAW ALIAS STUFF IN FAVOR OF const.VIEW
YAWALIAS = {
    # maybe this can be CODE_TO_SHORT
    'up': 'U',
    'down': 'D',
    'front': 'F',
    'left': 'L',
    'back': 'B',
    'right': 'R',
    'upfront': 'UF',
    'upback': 'UB',
    'upleft': 'UL',
    'upright': 'UR',
    'downfront': 'DF',
    'downback': 'DB',
    'downleft': 'DL',
    'downright': 'DR',
    'frontleft': 'FL',
    'frontright': 'FR',
    'backleft': 'BL',
    'backright': 'BR',
    'upfrontleft': 'UFL',
    'upfrontright': 'UFR',
    'upbackleft': 'UBL',
    'upbackright': 'UBR',
    'downfrontleft': 'DFL',
    'downfrontright': 'DFR',
    'downbackleft': 'DBL',
    'downbackright': 'DBR',
}

QUAL_EXCELLENT = 'excellent'
QUAL_GOOD = 'good'
QUAL_OK = 'ok'
QUAL_POOR = 'poor'
QUAL_JUNK = 'junk'
QUAL_UNKNOWN = 'UNKNOWN'

QUALITY_INT_TO_TEXT = OrderedDict(
    [
        (5, QUAL_EXCELLENT),
        (4, QUAL_GOOD),
        (3, QUAL_OK),
        (2, QUAL_POOR),
        # oops forgot 1. will be mapped to poor
        (0, QUAL_JUNK),
        (-1, QUAL_UNKNOWN),
    ]
)


QUALITY_TEXT_TO_INT = ut.invert_dict(QUALITY_INT_TO_TEXT)
QUALITY_INT_TO_TEXT[1] = QUAL_JUNK
# QUALITY_TEXT_TO_INTS      = ut.invert_dict(QUALITY_INT_TO_TEXT)
QUALITY_TEXT_TO_INTS = ut.group_items(
    list(QUALITY_INT_TO_TEXT.keys()), list(QUALITY_INT_TO_TEXT.values())
)
QUALITY_TEXT_TO_INTS[QUAL_UNKNOWN] = -1
QUALITY_INT_TO_TEXT[None] = QUALITY_INT_TO_TEXT[-1]


SEX_INT_TO_TEXT = {
    None: 'UNKNOWN NAME',
    -1: 'UNKNOWN SEX',
    0: 'Female',
    1: 'Male',
    2: 'INDETERMINATE SEX',
}
SEX_TEXT_TO_INT = ut.invert_dict(SEX_INT_TO_TEXT)


class PATH_NAMES(object):  # NOQA
    """ Path names for internal IBEIS database """

    sqldb = '_ibeis_database.sqlite3'
    sqlstaging = '_ibeis_staging.sqlite3'
    _ibsdb = '_ibsdb'
    cache = '_ibeis_cache'
    backups = '_ibeis_backups'
    logs = '_ibeis_logs'
    chips = 'chips'
    figures = 'figures'
    flann = 'flann'
    images = 'images'
    trees = 'trees'
    nets = 'nets'
    uploads = 'uploads'
    detectimg = 'detectimg'
    thumbs = 'thumbs'
    trashdir = 'trashed_images'
    distinctdir = 'distinctiveness_model'
    scorenormdir = 'scorenorm'
    smartpatrol = 'smart_patrol'
    # Query Results (chipmatch dirs)
    qres = 'qres_new'
    bigcache = 'qres_bigcache_new'


class REL_PATHS(object):  # NOQA
    """ all paths are relative to ibs.dbdir """

    _ibsdb = PATH_NAMES._ibsdb
    trashdir = PATH_NAMES.trashdir
    figures = join(_ibsdb, PATH_NAMES.figures)
    cache = join(_ibsdb, PATH_NAMES.cache)
    backups = join(_ibsdb, PATH_NAMES.backups)
    logs = join(_ibsdb, PATH_NAMES.logs)
    # chips       = join(_ibsdb, PATH_NAMES.chips)
    images = join(_ibsdb, PATH_NAMES.images)
    trees = join(_ibsdb, PATH_NAMES.trees)
    nets = join(_ibsdb, PATH_NAMES.nets)
    uploads = join(_ibsdb, PATH_NAMES.uploads)
    # All computed dirs live in <dbdir>/_ibsdb/_ibeis_cache
    chips = join(cache, PATH_NAMES.chips)
    thumbs = join(cache, PATH_NAMES.thumbs)
    flann = join(cache, PATH_NAMES.flann)
    qres = join(cache, PATH_NAMES.qres)
    bigcache = join(cache, PATH_NAMES.bigcache)
    distinctdir = join(cache, PATH_NAMES.distinctdir)


# Directories that should be excluded from copy operations
EXCLUDE_COPY_REL_DIRS = [
    REL_PATHS.chips,
    REL_PATHS.cache,
    REL_PATHS.backups,
    REL_PATHS.figures,
    REL_PATHS.nets,
    join(PATH_NAMES._ibsdb, '_ibeis_cache*'),
    # '_ibsdb/_ibeis_cache',
    '_ibsdb/chips',  # old path for caches
    './images',  # the hotspotter images dir
]


# TODO: Remove anything under this block completely


UNKNOWN_LBLANNOT_ROWID = 0
UNKNOWN_NAME_ROWID = 0
UNKNOWN_SPECIES_ROWID = 0
# Names normalized to the standard UNKNOWN_NAME
ACCEPTED_UNKNOWN_NAMES = set(['Unassigned'])

INDIVIDUAL_KEY = 'INDIVIDUAL_KEY'
SPECIES_KEY = 'SPECIES_KEY'
EMPTY_KEY = ''
UNKNOWN = '____'
KEY_DEFAULTS = {
    INDIVIDUAL_KEY: UNKNOWN,
    SPECIES_KEY: UNKNOWN,
}

# <UNFINISHED METADATA>
# We are letting wildbook do this metadata instead
# Define the special metadata for annotation

ROSEMARY_ANNOT_METADATA = [
    ('local_name', 'Local name:', str),
    ('sun', 'Sun:', ['FS', 'PS', 'NS']),
    ('wind', 'Wind:', ['NW', 'LW', 'MW', 'SW']),
    ('rain', 'Rain:', ['NR', 'LR', 'MR', 'HR']),
    ('cover', 'Cover:', float),
    ('grass', 'Grass:', ['less hf', 'less hk', 'less belly']),
    ('grass_color', 'Grass Colour:', ['B', 'BG', 'GB', 'G']),
    ('grass_species', 'Grass Species:', str),
    ('bush_type', 'Bush type:', ['OG', 'LB', 'MB', 'TB']),
    ('bit', 'Bit:', int),
    ('other_speceis', 'Other Species:', str),
]

# ROSEMARY_KEYS = utool.get_list_column(ROSEMARY_ANNOT_METADATA, 0)
# KEY_DEFAULTS.update(**{key: UNKNOWN for key in ROSEMARY_KEYS})
# </UNFINISHED METADATA>

BASE_DATABASE_VERSION = '0.0.0'

#################################################################
# DO NOT DELETE FROM THE TABLE LIST, THE DATABASE UPDATER WILL BREAK!!!
# THIS GOES FOR OLD AND DEPRICATED TABLENAMES AS WELL!!!
# TODO:
# What should happen is when they are depricated they should go into a
# depricated tablename structure with the relevant versions suffixed
#################################################################
# CORE DATABASE
AL_RELATION_TABLE = 'annotation_lblannot_relationship'
GA_RELATION_TABLE = 'annotgroup_annotation_relationship'
ANNOTGROUP_TABLE = 'annotgroups'
ANNOTATION_TABLE = 'annotations'
PART_TABLE = 'parts'
CHIP_TABLE = 'chips'
CONFIG_TABLE = 'configs'
CONTRIBUTOR_TABLE = 'contributors'
GSG_RELATION_TABLE = 'imageset_image_relationship'
IMAGESET_TABLE = 'imagesets'
FEATURE_TABLE = 'features'
FEATURE_WEIGHT_TABLE = 'feature_weights'
GL_RELATION_TABLE = 'image_lblimage_relationship'
IMAGE_TABLE = 'images'
LBLANNOT_TABLE = 'lblannot'
LBLIMAGE_TABLE = 'lblimage'
LBLTYPE_TABLE = 'keys'
METADATA_TABLE = 'metadata'
# Ugly move from name to names, need better way of versioning old table names
NAME_TABLE_v121 = 'name'
NAME_TABLE_v130 = 'names'
NAME_TABLE = NAME_TABLE_v130
ANNOTMATCH_TABLE = 'annotmatch'
SPECIES_TABLE = 'species'
RESIDUAL_TABLE = 'residuals'
VERSIONS_TABLE = 'versions'
#
PARTY_CONTRIB_RELATION_TABLE = 'party_contrib_relation'
PARTY_TABLE = 'party'
#################################################################
# STAGING DATABASE
REVIEW_TABLE = 'reviews'
TEST_TABLE = 'tests'
#################################################################


# DEPCACHE TABLENAMES
# CHIPTHUMB_TABLE = 'chipthumb'


UNKNOWN_PURPLE_RGBA255 = np.array((102, 0, 153, 255))
NAME_BLUE_RGBA255 = np.array((20, 20, 235, 255))
NAME_RED_RGBA255 = np.array((235, 20, 20, 255))
NEW_YELLOW_RGBA255 = np.array((235, 235, 20, 255))

UNKNOWN_PURPLE_RGBA01 = UNKNOWN_PURPLE_RGBA255 / 255.0
NAME_BLUE_RGBA01 = NAME_BLUE_RGBA255 / 255.0
NAME_RED_RGBA01 = NAME_RED_RGBA255 / 255.0
NEW_YELLOW_RGBA01 = NEW_YELLOW_RGBA255 / 255.0

EXEMPLAR_IMAGESETTEXT = '*Exemplars'
ALL_IMAGE_IMAGESETTEXT = '*All Images'
UNREVIEWED_IMAGE_IMAGESETTEXT = '*Undetected Images'
REVIEWED_IMAGE_IMAGESETTEXT = '*Reviewed Detections'
UNGROUPED_IMAGES_IMAGESETTEXT = '*Ungrouped Images'
SPECIAL_IMAGESET_LABELS = [
    EXEMPLAR_IMAGESETTEXT,
    ALL_IMAGE_IMAGESETTEXT,
    UNREVIEWED_IMAGE_IMAGESETTEXT,
    REVIEWED_IMAGE_IMAGESETTEXT,
    UNGROUPED_IMAGES_IMAGESETTEXT,
]
NEW_IMAGESET_IMAGESETTEXT = 'NEW IMAGESET'

# IMAGE_THUMB_SUFFIX = '_thumb.png'
# CHIP_THUMB_SUFFIX  = '_chip_thumb.png'
IMAGE_THUMB_SUFFIX = '_thumb.jpg'
IMAGE_BARE_THUMB_SUFFIX = '_thumb_bare.jpg'
CHIP_THUMB_SUFFIX = '_chip_thumb.jpg'


VS_EXEMPLARS_KEY = 'vs_exemplars'
INTRA_OCCUR_KEY = 'intra_occurrence'

HARD_NOTE_TAG = '<HARDCASE>'

# HACK
COMPUTER_NAME = ut.get_computer_name()
if COMPUTER_NAME in ['wbia.cs.uic.edu']:
    _DEFAULT_WILDBOOK_TARGET = 'lewa3'
elif COMPUTER_NAME in ['Leviathan']:
    _DEFAULT_WILDBOOK_TARGET = 'lewa3'
elif COMPUTER_NAME in ['maasai', 'quagga', 'xadmin-Nitro-AN515-51']:
    _DEFAULT_WILDBOOK_TARGET = 'quagga.princeton.edu'
else:
    _DEFAULT_WILDBOOK_TARGET = 'wbia'
WILDBOOK_TARGET = ut.get_argval(
    '--wildbook-target',
    type_=str,
    default=_DEFAULT_WILDBOOK_TARGET,
    help_='specify the Wildbook target deployment',
)


class ZIPPED_URLS(object):  # NOQA
    PZ_DISTINCTIVE = 'https://wildbookiarepository.azureedge.net/models/distinctivness_zebra_plains.zip'  # DEPRICATE
    GZ_DISTINCTIVE = 'https://wildbookiarepository.azureedge.net/models/distinctivness_zebra_grevys.zip'  # DEPRICATE

    PZ_MTEST = 'https://wildbookiarepository.azureedge.net/databases/PZ_MTEST.zip'
    NAUTS = 'https://wildbookiarepository.azureedge.net/databases/NAUT_test.zip'
    WDS = 'https://wildbookiarepository.azureedge.net/databases/wd_peter2.zip'
    DF_CURVRANK = (
        'https://wildbookiarepository.azureedge.net/databases/testdb_curvrank.zip'
    )
    ID_EXAMPLE = (
        'https://wildbookiarepository.azureedge.net/databases/testdb_identification.zip'
    )
    ORIENTATION = (
        'https://wildbookiarepository.azureedge.net/databases/testdb_orientation.zip'
    )
    K7_EXAMPLE = 'https://wildbookiarepository.azureedge.net/databases/testdb_kaggle7.zip'


# Turn off features at Lewa :(
SIMPLIFY_INTERFACE = (ut.get_computer_name() == 'wbia.cs.uic.edu') or ut.get_argflag(
    '--simplify'
)


# For candidacy document
DBNAME_ALIAS = {
    # 'NNP_MasterGIRM_core': 'NNP_GIRM'
    # 'NNP_MasterGIRM_core': 'GIRM',
    'NNP_MasterGIRM_core': 'GIRM',
    'PZ_Master1': 'PZ',
    'GZ_Master1': 'GZ',
    'GIRM_Master1': 'GIRM',
    'GZ_ALL': 'GZ',
}


class TEST_SPECIES(object):  # NOQA
    ZEB_PLAIN = 'zebra_plains'
    ZEB_GREVY = 'zebra_grevys'
    BEAR_POLAR = 'bear_polar'
    GIR_MASAI = 'giraffe_masai'


SPECIES_WITH_DETECTORS = (
    'cheetah',
    'dolphin_spotted',
    'dolphin_spotted+dorsal',
    'dolphin_spotted+fin_dorsal',
    'giraffe_masai',
    'giraffe_reticulated',
    'jaguar',
    'lynx',
    'manta_ray',
    'manta_ray_giant',
    'right_whale_head',
    'seadragon_leafy',
    'seadragon_leafy+head',
    'seadragon_weedy',
    'seadragon_weedy+head',
    'skunk_spotted',
    'turtle_green',
    'turtle_green+head',
    'turtle_hawksbill',
    'turtle_hawksbill+head',
    'turtle_oliveridley',
    'turtle_oliveridley+head',
    'turtle_sea',
    'turtle_sea+head',
    'whale_fluke',
    'whale_humpback',
    'whale_humpback+fin_dorsal',
    'zebra_grevys',
    'zebra_plains',
    'zebra_mountain',
    'whale_orca',
    'whale_orca+fin_dorsal',
    'leopard',
    'wild_dog',
    'wild_dog_dark',
    'wild_dog_light',
    'wild_dog_puppy',
    'wild_dog_standard',
    'wild_dog_tan',
    # Latin (Flukebook)
    'chelonioidea',
    'chelonioidea+head',
    'chelonia_mydas',
    'chelonia_mydas+head',
    'eretmochelys_imbricata',
    'eretmochelys_imbricata+head',
    'lepidochelys_olivacea',
    'lepidochelys_olivacea+head',
    'eubalaena_australis',
    'eubalaena_glacialis',
    'equus_quagga',
    'equus_grevyi',
    'equus_zebra',
    'giraffa_tippelskirchi',
    'giraffa_camelopardalis_reticulata',
    'mobula_birostris',
    'mobula_alfredi',
    'lynx_pardinus',
    'phycodurus_eques',
    'phycodurus_eques+head',
    'phyllopteryx_taeniolatus',
    'phyllopteryx_taeniolatus+head',
    'megaptera_novaeangliae',
    'physeter_macrocephalus',
    'physeter_macrocephalus+fin_dorsal',
    'spilogale_gracilis',
    'stenella_frontalis',
    'stenella_frontalis+dorsal',
    'stenella_frontalis+fin_dorsal',
    'orcinus_orca',
    'orcinus_orca+fin_dorsal',
    'panthera_pardus',
    'panthera_onca',
    'acinonyx_jubatus',
    'lycaon_pictus',
)


SPECIES_MAPPING = {
    'antelope': ('ANTEL', 'Antelope'),
    'airplane': ('PLANE', 'Airplane'),
    'bear_polar': ('PB', 'Polar Bear'),
    'bird': ('BIRD', 'Bird (Generic)'),
    'bird_generic': (None, 'bird'),
    'bicycle': ('BIKE', 'Bicycle'),
    'boat': ('BOAT', 'Boat'),
    'building': ('BUILDING', 'Building'),
    'bus': ('BUS', 'Bus'),
    'car': ('CAR', 'Car'),
    'camel': ('CAMEL', 'Camel'),
    'crane': ('CRANE', 'Crane'),
    'cheetah': ('CHTH', 'Cheetah'),
    'dog_wild': ('WD', 'Wild Dog'),
    'dolphin_fin': ('DF', 'Dolphin Fin'),
    'dolphin_bottlenose_fin': ('BDF', 'Bottlenose Dolphin Fin'),
    'dolphin_spotted': ('SDOLPH', 'Dolphin (Spotted)'),
    'domesticated_cat': ('DOMC', 'Cat (Domesticated)'),
    'domesticated_cow': ('DOMW', 'Cow (Domesticated)'),
    'domesticated_dog': ('DOMD', 'Dog (Domesticated)'),
    'domesticated_horse': ('DOMH', 'Horse (Domesticated)'),
    'domesticated_sheep': ('DOMS', 'Sheep (Domesticated)'),
    'cat_domestic': (None, 'domesticated_cat'),
    'cow_domestic': (None, 'domesticated_cow'),
    'dog_domestic': (None, 'domesticated_dog'),
    'horse_domestic': (None, 'domesticated_horse'),
    'sheep_domestic': (None, 'domesticated_sheep'),
    'elephant_savanna': ('ELEP', 'Elephant (Savanna)'),
    'elephant_savannah': (None, 'elephant_savanna'),
    'frog': ('FROG', 'Frog'),
    'gazelle': ('GAZ', 'Gazelle (Generic)'),
    'gazelle_generic': (None, 'gazelle'),
    'giraffe_masai': ('GIRM', 'Giraffe (Masai)'),
    'giraffe_massai': (None, 'giraffe_masai'),
    'giraffe_maasai': (None, 'giraffe_masai'),
    'giraffe_reticulated': ('GIR', 'Giraffe (Reticulated)'),
    'goat': ('GOAT', 'Goat (Generic)'),
    'goat_generic': (None, 'goat'),
    'grouper_nassau': ('NG', 'Nassau Grouper'),
    'hippopotamus': ('HIPPO', 'Hippopotamus'),
    'hyena': ('HYENA', 'Hyena'),
    'ignore': ('IGNORE', 'IGNORE'),
    'indistinct': ('INDISTINCT', 'INDISTINCT'),
    'jaguar': ('JAG', 'Jaguar'),
    'leopard': ('LOEP', 'Leopard'),
    'lion': ('LION', 'Lion'),
    'lionfish': ('LF', 'Lionfish'),
    'lynx': ('LYNX', 'Lynx'),
    'manta_alfredi': (None, 'mobula_alfredi'),
    'manta_birostris': (None, 'mobula_birostris'),
    'mobula_alfredi': ('MRR', 'Manta Ray (Reef)'),
    'mobula_birostris': ('MRG', 'Manta Ray (Giant)'),
    'manta_ray_giant': (None, 'mobula_birostris'),
    'manta_ray': (None, 'manta_ray_giant'),
    'morotcycle': ('MBIKE', 'Motorcycle'),
    'nautilus': ('NAUT', 'Nautilus'),
    'ostrich': ('OST', 'Ostrich'),
    'other': ('OTHER', 'Other'),
    'person': ('PERSON', 'Person'),
    'rhino_black': ('BRHINO', 'Rhino (Black)'),
    'rhino_white': ('WRHINO', 'Rhino (White)'),
    'seal_saimma_ringed': ('SEAL2', 'Seal (Siamaa Ringed)'),
    'seal_spotted': ('SEAL1', 'Seal (Spotted)'),
    'seadragon_leafy': ('SDL', 'Sea Dragon (Leafy)'),
    'seadragon_weedy': ('SDW', 'Sea Dragon (Weedy)'),
    'shark_sandtiger': ('STS', 'Sand Tiger Shark'),
    'skunk_spotted': ('SSKUNK', 'Skunk (Spotted)'),
    'snail': ('SNAIL', 'Snail'),
    'snow_leopard': ('SLEOP', 'Snow Leopard'),
    'tiger': ('TIGER', 'Tiger'),
    'turtle_sea': ('ST', 'Sea Turtle (Generic)'),
    'turtle_sea_generic': (None, 'turtle_sea'),
    'turtle_hawksbill': ('STHB', 'Sea Turtle (Hawksbill)'),
    'turtle_green': ('STG', 'Sea Turtle (Green)'),
    'train': ('TRAIN', 'Train'),
    'truck': ('TRUCK', 'Truck'),
    'toads_wyoming': ('WYTOAD', 'Toad (Wyoming)'),
    'unspecified': ('UNSPECIFIED', 'UNSPECIFIED'),
    'unspecified_animal': (None, 'unspecified'),
    'warthog': ('WART', 'Warthog'),
    'water_buffalo': ('BUFF', 'Water Buffalo'),
    'wildebeest': ('WB', 'Wildebeest'),
    'wild_dog_general': ('WDG', 'Wild Dog General'),
    'wild_dog_dark': ('WDD', 'Wild Dog Dark'),
    'wild_dog_tan': ('WDT', 'Wild Dog Tan'),
    'wild_dog_puppy': ('WDP', 'Wild Dog Puppy'),
    'wild_dog_standard': ('WDS', 'Wild Dog Standard'),
    'wild_dog+tail_general': ('WD+GEN', 'Wild Dog Tail General'),
    'wild_dog+tail_multi_black': ('WD+MB', 'Wild Dog Tail Multi-Black'),
    'wild_dog+tail_long_black': ('WD+LB', 'Wild Dog Tail Long Black'),
    'wild_dog+tail_long_white': ('WD+LW', 'Wild Dog Tail Long White'),
    'wild_dog+tail_long': ('WD+LONG', 'Wild Dog Tail Long'),
    'wild_dog+tail_double_black_brown': ('WD+DBB', 'Wild Dog Tail Double Black Brown'),
    'wild_dog+tail_double_black_white': ('WD+DBW', 'Wild Dog Tail Double Black White'),
    'wild_dog+tail_triple_black': ('WD+TB', 'Wild Dog Tail Triple Black'),
    'wild_dog+tail_short_black': ('WD+SB', 'Wild Dog Tail Short Black'),
    'wild_dog+tail_standard': ('WD+STD', 'Wild Dog Tail Standard'),
    'wild_dog': ('WD', 'Wild Dog'),
    'whale_fluke': ('WF', 'Whale Fluke'),
    'whale_humpback': ('HW', 'Humpback Whale'),
    'whale_shark': ('WS', 'Whale Shark'),
    'zebra_grevys': ('GZ', "Zebra (Grevy's)"),
    'zebra_hybrid': ('HZ', 'Zebra (Hybrid)'),
    'zebra_plains': ('PZ', 'Zebra (Plains)'),
    'zebra_mountain': ('MZ', 'Zebra (Mountain)'),
    UNKNOWN: ('UNKNOWN', 'UNSPECIFIED'),
}

PARTS_MAPPING = {
    'head': 'Head',
    'standard': 'Standard Tail',
    'short_black': 'Short Black Tail',
    'long_black': 'Long Black Tail',
    'double_black_brown': 'Double Black Brown Tail',
    'double_black_white': 'Double Black White Tail',
    'triple_black': 'Triple Black Tail',
    'long_white': 'Long White Tail',
    UNKNOWN: 'UNSPECIFIED,',
}


class _ConstHelper(type):
    """
    Adds code and nice constants to an integer version of a class
    """

    def __new__(cls, name, parents, dct):
        """
        cls = META_DECISION
        code_cls = META_DECISION_CODE
        """

        class CODE(object):
            pass

        class NICE(object):
            pass

        for key in dct.keys():
            if key.isupper():
                value = dct[key]
                if value is None or isinstance(value, int):
                    code = dct['INT_TO_CODE'][value]
                    nice = dct['INT_TO_NICE'][value]
                    setattr(CODE, key, code)
                    setattr(NICE, key, nice)

        dct['CODE'] = CODE
        dct['NICE'] = NICE
        # we need to call type.__new__ to complete the initialization
        return super(_ConstHelper, cls).__new__(cls, name, parents, dct)


@six.add_metaclass(_ConstHelper)
class EVIDENCE_DECISION(object):  # NOQA
    """
    TODO: change to EVIDENCE_DECISION / VISUAL_DECISION
    Enumerated types of review codes and texts

    Notes:
        Unreviewed: Not comparared yet.
        nomatch: Visually comparable and the different
        match: Visually comparable and the same
        notcomp: Not comparable means it is actually impossible to determine.
        unknown: means that it was reviewed, but we just can't figure it out.
    """

    UNREVIEWED = None
    NEGATIVE = 0
    POSITIVE = 1
    INCOMPARABLE = 2
    UNKNOWN = 3

    INT_TO_CODE = ut.odict(
        [
            (POSITIVE, 'match'),
            (NEGATIVE, 'nomatch'),
            (INCOMPARABLE, 'notcomp'),
            (UNKNOWN, 'unknown'),
            (UNREVIEWED, 'unreviewed'),
        ]
    )

    INT_TO_NICE = ut.odict(
        [
            (POSITIVE, 'Positive'),
            (NEGATIVE, 'Negative'),
            (INCOMPARABLE, 'Incomparable'),
            (UNKNOWN, 'Unknown'),
            (UNREVIEWED, 'Unreviewed'),
        ]
    )

    CODE_TO_NICE = ut.map_keys(INT_TO_CODE, INT_TO_NICE)
    CODE_TO_INT = ut.invert_dict(INT_TO_CODE)
    NICE_TO_CODE = ut.invert_dict(CODE_TO_NICE)
    NICE_TO_INT = ut.invert_dict(INT_TO_NICE)

    MATCH_CODE = CODE_TO_INT


@six.add_metaclass(_ConstHelper)
class META_DECISION(object):  # NOQA
    """
    Enumerated types of review codes and texts

    Notes:
        unreviewed: we dont have a meta decision
        same: we know this is the same animal through non-visual means
        diff: we know this is the different animal through non-visual means

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.constants import *  # NOQA
        >>> assert hasattr(META_DECISION, 'CODE')
        >>> assert hasattr(META_DECISION, 'NICE')
        >>> code1 = META_DECISION.INT_TO_CODE[META_DECISION.NULL]
        >>> code2 = META_DECISION.CODE.NULL
        >>> assert code1 == code2
        >>> nice1 = META_DECISION.INT_TO_NICE[META_DECISION.NULL]
        >>> nice2 = META_DECISION.NICE.NULL
        >>> assert nice1 == nice2
    """

    NULL = None
    DIFF = 0
    SAME = 1
    INT_TO_CODE = ut.odict([(NULL, 'null'), (DIFF, 'diff'), (SAME, 'same')])
    INT_TO_NICE = ut.odict([(NULL, 'NULL'), (DIFF, 'Different'), (SAME, 'Same')])
    CODE_TO_NICE = ut.map_keys(INT_TO_CODE, INT_TO_NICE)
    CODE_TO_INT = ut.invert_dict(INT_TO_CODE)
    NICE_TO_CODE = ut.invert_dict(CODE_TO_NICE)
    NICE_TO_INT = ut.invert_dict(INT_TO_NICE)


@six.add_metaclass(_ConstHelper)
class CONFIDENCE(object):
    UNKNOWN = None
    GUESSING = 1
    NOT_SURE = 2
    PRETTY_SURE = 3
    ABSOLUTELY_SURE = 4

    INT_TO_CODE = ut.odict(
        [
            (ABSOLUTELY_SURE, 'absolutely_sure'),
            (PRETTY_SURE, 'pretty_sure'),
            (NOT_SURE, 'not_sure'),
            (GUESSING, 'guessing'),
            (UNKNOWN, 'unspecified'),
        ]
    )

    INT_TO_NICE = ut.odict(
        [
            (ABSOLUTELY_SURE, 'Doubtless'),
            (PRETTY_SURE, 'Sure'),
            (NOT_SURE, 'Unsure'),
            (GUESSING, 'Guessing'),
            (UNKNOWN, 'Unspecified'),
        ]
    )

    CODE_TO_NICE = ut.map_keys(INT_TO_CODE, INT_TO_NICE)
    CODE_TO_INT = ut.invert_dict(INT_TO_CODE)
    NICE_TO_CODE = ut.invert_dict(CODE_TO_NICE)
    NICE_TO_INT = ut.invert_dict(INT_TO_NICE)


@six.add_metaclass(_ConstHelper)
class QUAL(object):
    EXCELLENT = 5
    GOOD = 4
    OK = 3
    POOR = 2
    JUNK = 1
    UNKNOWN = None

    INT_TO_CODE = ut.odict(
        [
            (EXCELLENT, 'excellent'),
            (GOOD, 'good'),
            (OK, 'ok'),
            (POOR, 'poor'),
            (JUNK, 'junk'),
            (UNKNOWN, 'unspecified'),
        ]
    )

    INT_TO_NICE = ut.odict(
        [
            (EXCELLENT, 'Excellent'),
            (GOOD, 'Good'),
            (OK, 'OK'),
            (POOR, 'Poor'),
            (JUNK, 'Junk'),
            (UNKNOWN, 'Unspecified'),
        ]
    )

    CODE_TO_NICE = ut.map_keys(INT_TO_CODE, INT_TO_NICE)
    CODE_TO_INT = ut.invert_dict(INT_TO_CODE)
    NICE_TO_CODE = ut.invert_dict(CODE_TO_NICE)
    NICE_TO_INT = ut.invert_dict(INT_TO_NICE)


@six.add_metaclass(_ConstHelper)
class VIEW(object):
    """
    categorical viewpoint using the faces of a Rhombicuboctahedron

    References:
        https://en.wikipedia.org/wiki/Rhombicuboctahedron
    """

    UNKNOWN = None
    R = 1
    FR = 2
    F = 3
    FL = 4
    L = 5
    BL = 6
    B = 7
    BR = 8

    U = 9
    UF = 10
    UB = 11
    UL = 12
    UR = 13
    UFL = 14
    UFR = 15
    UBL = 16
    UBR = 17

    D = 18
    DF = 19
    DB = 20
    DL = 21
    DR = 22
    DFL = 23
    DFR = 24
    DBL = 25
    DBR = 26

    INT_TO_CODE = ut.odict(
        [
            (UNKNOWN, 'unknown'),
            (R, 'right'),
            (FR, 'frontright'),
            (F, 'front'),
            (FL, 'frontleft'),
            (L, 'left'),
            (BL, 'backleft'),
            (B, 'back'),
            (BR, 'backright'),
            (U, 'up'),
            (UF, 'upfront'),
            (UB, 'upback'),
            (UL, 'upleft'),
            (UR, 'upright'),
            (UFL, 'upfrontleft'),
            (UFR, 'upfrontright'),
            (UBL, 'upbackleft'),
            (UBR, 'upbackright'),
            (D, 'down'),
            (DF, 'downfront'),
            (DB, 'downback'),
            (DL, 'downleft'),
            (DR, 'downright'),
            (DFL, 'downfrontleft'),
            (DFR, 'downfrontright'),
            (DBL, 'downbackleft'),
            (DBR, 'downbackright'),
        ]
    )

    INT_TO_NICE = ut.odict(
        [
            (UNKNOWN, 'Unknown'),
            (R, 'Right'),
            (FR, 'Front-Right'),
            (F, 'Front'),
            (FL, 'Front-Left'),
            (L, 'Left'),
            (BL, 'Back-Left'),
            (B, 'Back'),
            (BR, 'Back-Right'),
            (U, 'Up'),
            (UF, 'Up-Front'),
            (UB, 'Up-Back'),
            (UL, 'Up-Left'),
            (UR, 'Up-Right'),
            (UFL, 'Up-Front-Left'),
            (UFR, 'Up-Front-Right'),
            (UBL, 'Up-Back-Left'),
            (UBR, 'Up-Back-Right'),
            (D, 'Down'),
            (DF, 'Down-Front'),
            (DB, 'Down-Back'),
            (DL, 'Down-Left'),
            (DR, 'Down-Right'),
            (DFL, 'Down-Front-Left'),
            (DFR, 'Down-Front-Right'),
            (DBL, 'Down-Back-Left'),
            (DBR, 'Down-Back-Right'),
        ]
    )

    CODE_TO_NICE = ut.map_keys(INT_TO_CODE, INT_TO_NICE)
    CODE_TO_INT = ut.invert_dict(INT_TO_CODE)
    NICE_TO_CODE = ut.invert_dict(CODE_TO_NICE)
    NICE_TO_INT = ut.invert_dict(INT_TO_NICE)

    DIST = {
        # DIST 0 PAIRS
        (B, B): 0,
        (BL, BL): 0,
        (BR, BR): 0,
        (D, D): 0,
        (DB, DB): 0,
        (DBL, DBL): 0,
        (DBR, DBR): 0,
        (DF, DF): 0,
        (DFL, DFL): 0,
        (DFR, DFR): 0,
        (DL, DL): 0,
        (DR, DR): 0,
        (F, F): 0,
        (FL, FL): 0,
        (FR, FR): 0,
        (L, L): 0,
        (R, R): 0,
        (U, U): 0,
        (UB, UB): 0,
        (UBL, UBL): 0,
        (UBR, UBR): 0,
        (UF, UF): 0,
        (UFL, UFL): 0,
        (UFR, UFR): 0,
        (UL, UL): 0,
        (UR, UR): 0,
        # DIST 1 PAIRS
        (B, BL): 1,
        (B, BR): 1,
        (B, DB): 1,
        (B, DBL): 1,
        (B, DBR): 1,
        (B, UB): 1,
        (B, UBL): 1,
        (B, UBR): 1,
        (BL, DBL): 1,
        (BL, L): 1,
        (BL, UBL): 1,
        (BR, DBR): 1,
        (BR, R): 1,
        (BR, UBR): 1,
        (D, DB): 1,
        (D, DBL): 1,
        (D, DBR): 1,
        (D, DF): 1,
        (D, DFL): 1,
        (D, DFR): 1,
        (D, DL): 1,
        (D, DR): 1,
        (DB, DBL): 1,
        (DB, DBR): 1,
        (DBL, DL): 1,
        (DBL, L): 1,
        (DBR, DR): 1,
        (DBR, R): 1,
        (DF, DFL): 1,
        (DF, DFR): 1,
        (DF, F): 1,
        (DFL, DL): 1,
        (DFL, F): 1,
        (DFL, FL): 1,
        (DFL, L): 1,
        (DFR, DR): 1,
        (DFR, F): 1,
        (DFR, FR): 1,
        (DFR, R): 1,
        (DL, L): 1,
        (DR, R): 1,
        (F, FL): 1,
        (F, FR): 1,
        (F, UF): 1,
        (F, UFL): 1,
        (F, UFR): 1,
        (FL, L): 1,
        (FL, UFL): 1,
        (FR, R): 1,
        (FR, UFR): 1,
        (L, UBL): 1,
        (L, UFL): 1,
        (L, UL): 1,
        (R, UBR): 1,
        (R, UFR): 1,
        (R, UR): 1,
        (U, UB): 1,
        (U, UBL): 1,
        (U, UBR): 1,
        (U, UF): 1,
        (U, UFL): 1,
        (U, UFR): 1,
        (U, UL): 1,
        (U, UR): 1,
        (UB, UBL): 1,
        (UB, UBR): 1,
        (UBL, UL): 1,
        (UBR, UR): 1,
        (UF, UFL): 1,
        (UF, UFR): 1,
        (UFL, UL): 1,
        (UFR, UR): 1,
        # DIST 2 PAIRS
        (B, D): 2,
        (B, DL): 2,
        (B, DR): 2,
        (B, L): 2,
        (B, R): 2,
        (B, U): 2,
        (B, UL): 2,
        (B, UR): 2,
        (BL, BR): 2,
        (BL, D): 2,
        (BL, DB): 2,
        (BL, DBR): 2,
        (BL, DFL): 2,
        (BL, DL): 2,
        (BL, FL): 2,
        (BL, U): 2,
        (BL, UB): 2,
        (BL, UBR): 2,
        (BL, UFL): 2,
        (BL, UL): 2,
        (BR, D): 2,
        (BR, DB): 2,
        (BR, DBL): 2,
        (BR, DFR): 2,
        (BR, DR): 2,
        (BR, FR): 2,
        (BR, U): 2,
        (BR, UB): 2,
        (BR, UBL): 2,
        (BR, UFR): 2,
        (BR, UR): 2,
        (D, F): 2,
        (D, FL): 2,
        (D, FR): 2,
        (D, L): 2,
        (D, R): 2,
        (DB, DF): 2,
        (DB, DFL): 2,
        (DB, DFR): 2,
        (DB, DL): 2,
        (DB, DR): 2,
        (DB, L): 2,
        (DB, R): 2,
        (DB, UB): 2,
        (DB, UBL): 2,
        (DB, UBR): 2,
        (DBL, DBR): 2,
        (DBL, DF): 2,
        (DBL, DFL): 2,
        (DBL, DFR): 2,
        (DBL, DR): 2,
        (DBL, FL): 2,
        (DBL, UB): 2,
        (DBL, UBL): 2,
        (DBL, UBR): 2,
        (DBL, UFL): 2,
        (DBL, UL): 2,
        (DBR, DF): 2,
        (DBR, DFL): 2,
        (DBR, DFR): 2,
        (DBR, DL): 2,
        (DBR, FR): 2,
        (DBR, UB): 2,
        (DBR, UBL): 2,
        (DBR, UBR): 2,
        (DBR, UFR): 2,
        (DBR, UR): 2,
        (DF, DL): 2,
        (DF, DR): 2,
        (DF, FL): 2,
        (DF, FR): 2,
        (DF, L): 2,
        (DF, R): 2,
        (DF, UF): 2,
        (DF, UFL): 2,
        (DF, UFR): 2,
        (DFL, DFR): 2,
        (DFL, DR): 2,
        (DFL, FR): 2,
        (DFL, UBL): 2,
        (DFL, UF): 2,
        (DFL, UFL): 2,
        (DFL, UFR): 2,
        (DFL, UL): 2,
        (DFR, DL): 2,
        (DFR, FL): 2,
        (DFR, UBR): 2,
        (DFR, UF): 2,
        (DFR, UFL): 2,
        (DFR, UFR): 2,
        (DFR, UR): 2,
        (DL, DR): 2,
        (DL, F): 2,
        (DL, FL): 2,
        (DL, UBL): 2,
        (DL, UFL): 2,
        (DL, UL): 2,
        (DR, F): 2,
        (DR, FR): 2,
        (DR, UBR): 2,
        (DR, UFR): 2,
        (DR, UR): 2,
        (F, L): 2,
        (F, R): 2,
        (F, U): 2,
        (F, UL): 2,
        (F, UR): 2,
        (FL, FR): 2,
        (FL, U): 2,
        (FL, UBL): 2,
        (FL, UF): 2,
        (FL, UFR): 2,
        (FL, UL): 2,
        (FR, U): 2,
        (FR, UBR): 2,
        (FR, UF): 2,
        (FR, UFL): 2,
        (FR, UR): 2,
        (L, U): 2,
        (L, UB): 2,
        (L, UF): 2,
        (R, U): 2,
        (R, UB): 2,
        (R, UF): 2,
        (UB, UF): 2,
        (UB, UFL): 2,
        (UB, UFR): 2,
        (UB, UL): 2,
        (UB, UR): 2,
        (UBL, UBR): 2,
        (UBL, UF): 2,
        (UBL, UFL): 2,
        (UBL, UFR): 2,
        (UBL, UR): 2,
        (UBR, UF): 2,
        (UBR, UFL): 2,
        (UBR, UFR): 2,
        (UBR, UL): 2,
        (UF, UL): 2,
        (UF, UR): 2,
        (UFL, UFR): 2,
        (UFL, UR): 2,
        (UFR, UL): 2,
        (UL, UR): 2,
        # DIST 3 PAIRS
        (B, DF): 3,
        (B, DFL): 3,
        (B, DFR): 3,
        (B, FL): 3,
        (B, FR): 3,
        (B, UF): 3,
        (B, UFL): 3,
        (B, UFR): 3,
        (BL, DF): 3,
        (BL, DFR): 3,
        (BL, DR): 3,
        (BL, F): 3,
        (BL, R): 3,
        (BL, UF): 3,
        (BL, UFR): 3,
        (BL, UR): 3,
        (BR, DF): 3,
        (BR, DFL): 3,
        (BR, DL): 3,
        (BR, F): 3,
        (BR, L): 3,
        (BR, UF): 3,
        (BR, UFL): 3,
        (BR, UL): 3,
        (D, UB): 3,
        (D, UBL): 3,
        (D, UBR): 3,
        (D, UF): 3,
        (D, UFL): 3,
        (D, UFR): 3,
        (D, UL): 3,
        (D, UR): 3,
        (DB, F): 3,
        (DB, FL): 3,
        (DB, FR): 3,
        (DB, U): 3,
        (DB, UFL): 3,
        (DB, UFR): 3,
        (DB, UL): 3,
        (DB, UR): 3,
        (DBL, F): 3,
        (DBL, FR): 3,
        (DBL, R): 3,
        (DBL, U): 3,
        (DBL, UF): 3,
        (DBL, UR): 3,
        (DBR, F): 3,
        (DBR, FL): 3,
        (DBR, L): 3,
        (DBR, U): 3,
        (DBR, UF): 3,
        (DBR, UL): 3,
        (DF, U): 3,
        (DF, UBL): 3,
        (DF, UBR): 3,
        (DF, UL): 3,
        (DF, UR): 3,
        (DFL, R): 3,
        (DFL, U): 3,
        (DFL, UB): 3,
        (DFL, UR): 3,
        (DFR, L): 3,
        (DFR, U): 3,
        (DFR, UB): 3,
        (DFR, UL): 3,
        (DL, FR): 3,
        (DL, R): 3,
        (DL, U): 3,
        (DL, UB): 3,
        (DL, UBR): 3,
        (DL, UF): 3,
        (DL, UFR): 3,
        (DR, FL): 3,
        (DR, L): 3,
        (DR, U): 3,
        (DR, UB): 3,
        (DR, UBL): 3,
        (DR, UF): 3,
        (DR, UFL): 3,
        (F, UB): 3,
        (F, UBL): 3,
        (F, UBR): 3,
        (FL, R): 3,
        (FL, UB): 3,
        (FL, UBR): 3,
        (FL, UR): 3,
        (FR, L): 3,
        (FR, UB): 3,
        (FR, UBL): 3,
        (FR, UL): 3,
        (L, UBR): 3,
        (L, UFR): 3,
        (L, UR): 3,
        (R, UBL): 3,
        (R, UFL): 3,
        (R, UL): 3,
        # DIST 4 PAIRS
        (B, F): 4,
        (BL, FR): 4,
        (BR, FL): 4,
        (D, U): 4,
        (DB, UF): 4,
        (DBL, UFR): 4,
        (DBR, UFL): 4,
        (DF, UB): 4,
        (DFL, UBR): 4,
        (DFR, UBL): 4,
        (DL, UR): 4,
        (DR, UL): 4,
        (L, R): 4,
        # UNDEFINED DIST PAIRS
        (B, UNKNOWN): None,
        (BL, UNKNOWN): None,
        (BR, UNKNOWN): None,
        (D, UNKNOWN): None,
        (DB, UNKNOWN): None,
        (DBL, UNKNOWN): None,
        (DBR, UNKNOWN): None,
        (DF, UNKNOWN): None,
        (DFL, UNKNOWN): None,
        (DFR, UNKNOWN): None,
        (DL, UNKNOWN): None,
        (DR, UNKNOWN): None,
        (F, UNKNOWN): None,
        (FL, UNKNOWN): None,
        (FR, UNKNOWN): None,
        (L, UNKNOWN): None,
        (R, UNKNOWN): None,
        (U, UNKNOWN): None,
        (UB, UNKNOWN): None,
        (UBL, UNKNOWN): None,
        (UBR, UNKNOWN): None,
        (UF, UNKNOWN): None,
        (UFL, UNKNOWN): None,
        (UFR, UNKNOWN): None,
        (UL, UNKNOWN): None,
        (UNKNOWN, B): None,
        (UNKNOWN, BL): None,
        (UNKNOWN, BR): None,
        (UNKNOWN, D): None,
        (UNKNOWN, DB): None,
        (UNKNOWN, DBL): None,
        (UNKNOWN, DBR): None,
        (UNKNOWN, DF): None,
        (UNKNOWN, DFL): None,
        (UNKNOWN, DFR): None,
        (UNKNOWN, DL): None,
        (UNKNOWN, DR): None,
        (UNKNOWN, F): None,
        (UNKNOWN, FL): None,
        (UNKNOWN, FR): None,
        (UNKNOWN, L): None,
        (UNKNOWN, R): None,
        (UNKNOWN, U): None,
        (UNKNOWN, UB): None,
        (UNKNOWN, UBL): None,
        (UNKNOWN, UBR): None,
        (UNKNOWN, UF): None,
        (UNKNOWN, UFL): None,
        (UNKNOWN, UFR): None,
        (UNKNOWN, UL): None,
        (UNKNOWN, UR): None,
        (UR, UNKNOWN): None,
        (UNKNOWN, UNKNOWN): None,
    }
    # make distance symmetric
    for (f1, f2), d in list(DIST.items()):
        DIST[(f2, f1)] = d



# -*- coding: utf-8 -*-
"""
CommandLine:
    # Autogenerate ImageSet Functions
    # key should be the table name
    # the write flag makes a file, but dont use that
    python -m wbia.templates.template_generator --key imageset_image_relationship --onlyfn
"""
import logging
from wbia import constants as const
from wbia.control import accessor_decors
from wbia.control.controller_inject import make_ibs_register_decorator
import utool as ut

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


CLASS_INJECT_KEY, register_ibs_method = make_ibs_register_decorator(__name__)


@register_ibs_method
def delete_empty_imgsetids(ibs):
    """Removes imagesets without images

    Args:
        ibs (IBEISController):  wbia controller object

    CommandLine:
        python -m wbia.control.manual_gsgrelate_funcs --test-delete_empty_imgsetids

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_gsgrelate_funcs import *  # NOQA
        >>> import wbia
        >>> # build test data
        >>> ibs = wbia.opendb('testdb1')
        >>> # execute function
        >>> result = ibs.delete_empty_imgsetids()
        >>> # verify results
        >>> print(result)
    """
    imgsetid_list = ibs.get_valid_imgsetids(min_num_gids=0)
    nGids_list = ibs.get_imageset_num_gids(imgsetid_list)
    is_invalid = [nGids == 0 for nGids in nGids_list]
    invalid_imgsetids = ut.compress(imgsetid_list, is_invalid)
    ibs.delete_imagesets(invalid_imgsetids)


@register_ibs_method
@accessor_decors.getter_1to1
def get_image_gsgrids(ibs, gid_list):
    """
    Returns:
        list_ (list):  a list of imageset-image-relationship rowids for each imageid"""
    # TODO: Group type
    params_iter = ((gid,) for gid in gid_list)
    where_clause = 'image_rowid=?'
    # list of relationships for each image
    gsgrids_list = ibs.db.get_where(
        const.GSG_RELATION_TABLE,
        ('gsgr_rowid',),
        params_iter,
        where_clause,
        unpack_scalars=False,
    )
    return gsgrids_list


@register_ibs_method
@accessor_decors.deleter
@accessor_decors.cache_invalidator(const.IMAGESET_TABLE, ['percent_imgs_reviewed_str'])
@accessor_decors.cache_invalidator(
    const.IMAGESET_TABLE, ['percent_names_with_exemplar_str']
)
@accessor_decors.cache_invalidator(
    const.IMAGESET_TABLE, ['percent_annotmatch_reviewed_str']
)
def delete_gsgr_imageset_relations(ibs, imgsetid_list):
    """ Removes relationship between input imagesets and all images """
    ibs.db.delete(const.GSG_RELATION_TABLE, imgsetid_list, id_colname='imageset_rowid')


@register_ibs_method
@accessor_decors.deleter
@accessor_decors.cache_invalidator(const.IMAGESET_TABLE, ['percent_imgs_reviewed_str'])
@accessor_decors.cache_invalidator(
    const.IMAGESET_TABLE, ['percent_names_with_exemplar_str']
)
@accessor_decors.cache_invalidator(
    const.IMAGESET_TABLE, ['percent_annotmatch_reviewed_str']
)
def delete_gsgr_image_relations(ibs, gid_list):
    """ Removes relationship between input images and all imagesets """
    ibs.db.delete(const.GSG_RELATION_TABLE, gid_list, id_colname='image_rowid')


@register_ibs_method
@accessor_decors.deleter
@accessor_decors.cache_invalidator(const.IMAGESET_TABLE, ['image_rowids'], rowidx=1)
@accessor_decors.cache_invalidator(const.IMAGESET_TABLE, ['percent_imgs_reviewed_str'])
@accessor_decors.cache_invalidator(
    const.IMAGESET_TABLE, ['percent_names_with_exemplar_str']
)
@accessor_decors.cache_invalidator(
    const.IMAGESET_TABLE, ['percent_annotmatch_reviewed_str']
)
def unrelate_images_and_imagesets(ibs, gid_list, imgsetid_list):
    """
    Seems to unrelate specific image imageset pairs

    Args:
        ibs (IBEISController):  wbia controller object
        gid_list (list):
        imgsetid_list (list):

    Returns:
        list: gids_list

    CommandLine:
        python -m wbia.control.manual_gsgrelate_funcs --test-unrelate_images_and_imagesets
        python -c "import utool; print(utool.auto_docstr('wbia.control.manual_gsgrelate_funcs', 'delete_gsgr_image_relations'))"

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_gsgrelate_funcs import *  # NOQA
        >>> import wbia
        >>> # build test data
        >>> ibs = wbia.opendb('testdb1')
        >>> # Reset and compute imagesets
        >>> ibs.delete_all_imagesets()
        >>> ibs.compute_occurrences(config={'use_gps': False, 'seconds_thresh': 600})
        >>> imgsetid_list = ibs.get_valid_imgsetids()
        >>> gids_list = ibs.get_imageset_gids(imgsetid_list)
        >>> assert len(imgsetid_list) == 2, 'bad len %r' % (len(imgsetid_list),)
        >>> assert len(gids_list) == 2, 'bad len %r' % (len(gids_list),)
        >>> assert len(gids_list[0]) == 7, 'bad len %r' % (len(gids_list[0]),)
        >>> assert len(gids_list[1]) == 6, 'bad len %r' % (len(gids_list[1]),)
        >>> # Add imageset 2 gids to imageset 1 so an image belongs to multiple imagesets
        >>> imgset2_gids = gids_list[1][0:1]
        >>> imgset1_imgsetids = imgsetid_list[0:1]
        >>> ibs.add_image_relationship(imgset2_gids, imgset1_imgsetids)
        >>> # Now delete the image from the imageset 2
        >>> imgset2_imgsetids = imgsetid_list[1:2]
        >>> # execute function
        >>> ibs.unrelate_images_and_imagesets(imgset2_gids, imgset2_imgsetids)
        >>> # verify results
        >>> ibs.print_egpairs_table()
        >>> imgsetid_list_ = ibs.get_valid_imgsetids()
        >>> gids_list_ = ibs.get_imageset_gids(imgsetid_list_)
        >>> result = str(gids_list_)
        >>> print(result)
        >>> # imgset2_gids should now only be in imageset1
        >>> assert imgset2_gids[0] in gids_list_[0], 'imgset2_gids should now only be in imageset1'
        >>> assert imgset2_gids[0] not in gids_list_[1], 'imgset2_gids should now only be in imageset1'
    """
    # WHAT IS THIS FUNCTION? FIXME CALLS WEIRD FUNCTION
    if ut.VERBOSE:
        logger.info("[ibs] deleting %r image's imageset ids" % len(gid_list))
    gsgrid_list = ut.flatten(
        ibs.get_imageset_gsgrids(imgsetid_list=imgsetid_list, gid_list=gid_list)
    )
    ibs.db.delete_rowids(const.GSG_RELATION_TABLE, gsgrid_list)


# GETTERS::GSG_RELATION_TABLE


@register_ibs_method
@accessor_decors.getter_1to1
def get_gsgr_rowid_from_superkey(ibs, gid_list, imgsetid_list):
    """
    Returns:
        gsgrid_list (list):  eg-relate-ids from info constrained to be unique (imgsetid, gid)"""
    colnames = ('image_rowid',)
    params_iter = zip(gid_list, imgsetid_list)
    where_clause = 'image_rowid=? AND imageset_rowid=?'
    gsgrid_list = ibs.db.get_where(
        const.GSG_RELATION_TABLE, colnames, params_iter, where_clause
    )
    return gsgrid_list


@register_ibs_method
@accessor_decors.adder
@accessor_decors.cache_invalidator(const.IMAGESET_TABLE, ['image_rowids'], rowidx=1)
@accessor_decors.cache_invalidator(
    const.IMAGESET_TABLE, ['percent_imgs_reviewed_str'], rowidx=1
)
@accessor_decors.cache_invalidator(
    const.IMAGESET_TABLE, ['percent_names_with_exemplar_str'], rowidx=1
)
@accessor_decors.cache_invalidator(
    const.IMAGESET_TABLE, ['percent_annotmatch_reviewed_str'], rowidx=1
)
def add_image_relationship(ibs, gid_list, imgsetid_list):
    """ Adds a relationship between an image and and imageset """
    colnames = (
        'image_rowid',
        'imageset_rowid',
    )
    params_iter = list(zip(gid_list, imgsetid_list))
    get_rowid_from_superkey = ibs.get_gsgr_rowid_from_superkey
    superkey_paramx = (0, 1)
    gsgrid_list = ibs.db.add_cleanly(
        const.GSG_RELATION_TABLE,
        colnames,
        params_iter,
        get_rowid_from_superkey,
        superkey_paramx,
    )
    return gsgrid_list



# -*- coding: utf-8 -*-
"""
Dependencies: flask, tornado
"""
import logging
from flask import request, make_response, current_app
from wbia.control import controller_inject
from wbia.web import appfuncs as appf
import utool as ut
import vtool as vt

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')

register_route = controller_inject.get_wbia_flask_route(__name__)


@register_route('/ajax/cookie/', methods=['GET'])
def set_cookie(**kwargs):
    response = make_response('true')
    response.set_cookie(request.args['name'], request.args['value'])
    logger.info(
        '[web] Set Cookie: %r -> %r' % (request.args['name'], request.args['value'])
    )
    return response


def _resize_src(image, resize=False, **kwargs):
    # Load image
    if resize is None:
        image_src = appf.embed_image_html(image, target_width=None, target_height=None)
    elif resize:
        image = appf.resize_via_web_parameters(image)
        image_src = appf.embed_image_html(image, target_width=None, target_height=None)
    else:
        image_src = appf.embed_image_html(image)

    return image_src


@register_route('/ajax/image/src/<gid>/', methods=['GET'])
def image_src(gid=None, thumbnail=False, ibs=None, **kwargs):
    if ibs is None:
        ibs = current_app.ibs

    gid = int(gid)
    gpath = None

    if 'thumbsize' not in kwargs:
        kwargs['thumbsize'] = max(int(appf.TARGET_WIDTH), int(appf.TARGET_HEIGHT))

    if 'draw_annots' not in kwargs:
        kwargs['draw_annots'] = False

    if thumbnail:
        try:
            gpath = ibs.get_image_thumbpath(gid, ensure_paths=True, **kwargs)
            orient = ibs.get_image_orientation(gid)
            image = vt.imread(gpath, orient=orient)
            h, w = image.shape[:2]
            assert h > 0, 'Invalid image thumbnail'
            assert w > 0, 'Invalid image thumbnail'
        except AssertionError:
            gpath = None

    if gpath is None:
        gpath = ibs.get_image_paths(gid)

    image_src = image_src_path(gpath, **kwargs)
    return image_src


def image_src_path(gpath, orient='auto', **kwargs):
    image = vt.imread(gpath, orient=orient)
    image_src = _resize_src(image, **kwargs)
    return image_src


@register_route('/ajax/annot/src/<aid>/', methods=['GET'])
def annotation_src(aid=None, ibs=None, **kwargs):
    if ibs is None:
        ibs = current_app.ibs

    if 'dim_size' not in kwargs:
        kwargs['dim_size'] = max(int(appf.TARGET_WIDTH), int(appf.TARGET_HEIGHT))
    image = ibs.get_annot_chips(aid, config2_=kwargs)

    # image_src = _resize_src(image, **kwargs)
    image_src = appf.embed_image_html(image, target_height=300)
    return image_src


@register_route('/ajax/background/src/<aid>/', methods=['GET'])
def probchip_src(aid=None, ibs=None, **kwargs):
    import cv2

    if ibs is None:
        ibs = current_app.ibs

    if 'dim_size' not in kwargs:
        kwargs['dim_size'] = max(int(appf.TARGET_WIDTH), int(appf.TARGET_HEIGHT))

    image_filepath = ibs.get_annot_probchip_fpath(aid, config2_=kwargs)
    image = cv2.imread(image_filepath)
    x, y, w, h = ibs.get_annot_bboxes(aid)

    image = cv2.resize(image, (w, h))
    # image_src = _resize_src(image, **kwargs)
    image_src = appf.embed_image_html(image, target_height=300)
    return image_src


@register_route('/ajax/part/src/<part_rowid>/', methods=['GET'])
def part_src(part_rowid, **kwargs):
    ibs = current_app.ibs
    if 'dim_size' not in kwargs:
        kwargs['dim_size'] = max(int(appf.TARGET_WIDTH), int(appf.TARGET_HEIGHT))
    image = ibs.get_part_chips(part_rowid, config2_=kwargs)
    image_src = appf.embed_image_html(image, target_height=300)
    return image_src



# -*- coding: utf-8 -*-
import logging
import utool as ut
from os.path import join, dirname, basename

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


def get_injured_sharks():
    """
    >>> from wbia.scripts.getshark import *  # NOQA
    """
    import requests

    url = 'http://www.whaleshark.org/getKeywordImages.jsp'
    resp = requests.get(url)
    assert resp.status_code == 200
    keywords = resp.json()['keywords']
    key_list = ut.take_column(keywords, 'indexName')
    key_to_nice = {k['indexName']: k['readableName'] for k in keywords}

    injury_patterns = [
        'injury',
        'net',
        'hook',
        'trunc',
        'damage',
        'scar',
        'nicks',
        'bite',
    ]

    injury_keys = [
        key for key in key_list if any([pat in key for pat in injury_patterns])
    ]
    noninjury_keys = ut.setdiff(key_list, injury_keys)
    injury_nice = ut.lmap(lambda k: key_to_nice[k], injury_keys)  # NOQA
    noninjury_nice = ut.lmap(lambda k: key_to_nice[k], noninjury_keys)  # NOQA
    key_list = injury_keys

    keyed_images = {}
    for key in ut.ProgIter(key_list, lbl='reading index', bs=True):
        key_url = url + '?indexName={indexName}'.format(indexName=key)
        key_resp = requests.get(key_url)
        assert key_resp.status_code == 200
        key_imgs = key_resp.json()['images']
        keyed_images[key] = key_imgs

    key_hist = {key: len(imgs) for key, imgs in keyed_images.items()}
    key_hist = ut.sort_dict(key_hist, 'vals')
    logger.info(ut.repr3(key_hist))
    nice_key_hist = ut.map_dict_keys(lambda k: key_to_nice[k], key_hist)
    nice_key_hist = ut.sort_dict(nice_key_hist, 'vals')
    logger.info(ut.repr3(nice_key_hist))

    key_to_urls = {key: ut.take_column(vals, 'url') for key, vals in keyed_images.items()}
    overlaps = {}
    import itertools

    overlap_img_list = []
    for k1, k2 in itertools.combinations(key_to_urls.keys(), 2):
        overlap_imgs = ut.isect(key_to_urls[k1], key_to_urls[k2])
        num_overlap = len(overlap_imgs)
        overlaps[(k1, k2)] = num_overlap
        overlaps[(k1, k1)] = len(key_to_urls[k1])
        if num_overlap > 0:
            # logger.info('[%s][%s], overlap=%r' % (k1, k2, num_overlap))
            overlap_img_list.extend(overlap_imgs)

    all_img_urls = list(set(ut.flatten(key_to_urls.values())))
    num_all = len(all_img_urls)  # NOQA
    logger.info('num_all = %r' % (num_all,))

    # Determine super-categories
    categories = ['nicks', 'scar', 'trunc']

    # Force these keys into these categories
    key_to_cat = {'scarbite': 'other_injury'}

    cat_to_keys = ut.ddict(list)

    for key in key_to_urls.keys():
        flag = 1
        if key in key_to_cat:
            cat = key_to_cat[key]
            cat_to_keys[cat].append(key)
            continue
        for cat in categories:
            if cat in key:
                cat_to_keys[cat].append(key)
                flag = 0
        if flag:
            cat = 'other_injury'
            cat_to_keys[cat].append(key)

    cat_urls = ut.ddict(list)
    for cat, keys in cat_to_keys.items():
        for key in keys:
            cat_urls[cat].extend(key_to_urls[key])

    cat_hist = {}
    for cat in list(cat_urls.keys()):
        cat_urls[cat] = list(set(cat_urls[cat]))
        cat_hist[cat] = len(cat_urls[cat])

    logger.info(ut.repr3(cat_to_keys))
    logger.info(ut.repr3(cat_hist))

    key_to_cat = dict([(val, key) for key, vals in cat_to_keys.items() for val in vals])

    # ingestset = {
    #    '__class__': 'ImageSet',
    #    'images': ut.ddict(dict)
    # }
    # for key, key_imgs in keyed_images.items():
    #    for imgdict in key_imgs:
    #        url = imgdict['url']
    #        encid = imgdict['correspondingEncounterNumber']
    #        # Make structure
    #        encdict = encounters[encid]
    #        encdict['__class__'] = 'Encounter'
    #        imgdict = ut.delete_keys(imgdict.copy(), ['correspondingEncounterNumber'])
    #        imgdict['__class__'] = 'Image'
    #        cat = key_to_cat[key]
    #        annotdict = {'relative_bbox': [.01, .01, .98, .98], 'tags': [cat, key]}
    #        annotdict['__class__'] = 'Annotation'

    #        # Ensure structures exist
    #        encdict['images'] = encdict.get('images', [])
    #        imgdict['annots'] = imgdict.get('annots', [])

    #        # Add an image to this encounter
    #        encdict['images'].append(imgdict)
    #        # Add an annotation to this image
    #        imgdict['annots'].append(annotdict)

    # # http://springbreak.wildbook.org/rest/org.ecocean.Encounter/1111
    # get_enc_url = 'http://www.whaleshark.org/rest/org.ecocean.Encounter/%s' % (encid,)
    # resp = requests.get(get_enc_url)
    # logger.info(ut.repr3(encdict))
    # logger.info(ut.repr3(encounters))

    # Download the files to the local disk
    # fpath_list =

    all_urls = ut.unique(
        ut.take_column(
            ut.flatten(
                ut.dict_subset(keyed_images, ut.flatten(cat_to_keys.values())).values()
            ),
            'url',
        )
    )

    dldir = ut.truepath('~/tmpsharks')
    from os.path import commonprefix, basename  # NOQA

    prefix = commonprefix(all_urls)
    suffix_list = [url_[len(prefix) :] for url_ in all_urls]
    fname_list = [suffix.replace('/', '--') for suffix in suffix_list]

    fpath_list = []
    for url, fname in ut.ProgIter(
        zip(all_urls, fname_list), lbl='downloading imgs', freq=1
    ):
        fpath = ut.grab_file_url(url, download_dir=dldir, fname=fname, verbose=False)
        fpath_list.append(fpath)

    # Make sure we keep orig info
    # url_to_keys = ut.ddict(list)
    url_to_info = ut.ddict(dict)
    for key, imgdict_list in keyed_images.items():
        for imgdict in imgdict_list:
            url = imgdict['url']
            info = url_to_info[url]
            for k, v in imgdict.items():
                info[k] = info.get(k, [])
                info[k].append(v)
            info['keys'] = info.get('keys', [])
            info['keys'].append(key)
            # url_to_keys[url].append(key)

    info_list = ut.take(url_to_info, all_urls)
    for info in info_list:
        if len(set(info['correspondingEncounterNumber'])) > 1:
            assert False, 'url with two different encounter nums'
    # Combine duplicate tags

    hashid_list = [
        ut.get_file_uuid(fpath_, stride=8) for fpath_ in ut.ProgIter(fpath_list, bs=True)
    ]
    groupxs = ut.group_indices(hashid_list)[1]

    # Group properties by duplicate images
    # groupxs = [g for g in groupxs if len(g) > 1]
    fpath_list_ = ut.take_column(ut.apply_grouping(fpath_list, groupxs), 0)
    url_list_ = ut.take_column(ut.apply_grouping(all_urls, groupxs), 0)
    info_list_ = [
        ut.map_dict_vals(ut.flatten, ut.dict_accum(*info_))
        for info_ in ut.apply_grouping(info_list, groupxs)
    ]

    encid_list_ = [
        ut.unique(info_['correspondingEncounterNumber'])[0] for info_ in info_list_
    ]
    keys_list_ = [ut.unique(info_['keys']) for info_ in info_list_]
    cats_list_ = [ut.unique(ut.take(key_to_cat, keys)) for keys in keys_list_]

    clist = ut.ColumnLists(
        {
            'gpath': fpath_list_,
            'url': url_list_,
            'encid': encid_list_,
            'key': keys_list_,
            'cat': cats_list_,
        }
    )

    # for info_ in ut.apply_grouping(info_list, groupxs):
    #    info = ut.dict_accum(*info_)
    #    info = ut.map_dict_vals(ut.flatten, info)
    #    x = ut.unique(ut.flatten(ut.dict_accum(*info_)['correspondingEncounterNumber']))
    #    if len(x) > 1:
    #        info = info.copy()
    #        del info['keys']
    #        logger.info(ut.repr3(info))

    flags = ut.lmap(ut.fpath_has_imgext, clist['gpath'])
    clist = clist.compress(flags)

    import wbia

    ibs = wbia.opendb('WS_Injury', allow_newdir=True)

    gid_list = ibs.add_images(clist['gpath'])
    clist['gid'] = gid_list

    failed_flags = ut.flag_None_items(clist['gid'])
    logger.info('# failed %s' % (sum(failed_flags),))
    passed_flags = ut.not_list(failed_flags)
    clist = clist.compress(passed_flags)
    ut.assert_all_not_None(clist['gid'])
    # ibs.get_image_uris_original(clist['gid'])
    ibs.set_image_uris_original(clist['gid'], clist['url'], overwrite=True)

    # ut.zipflat(clist['cat'], clist['key'])
    if False:
        # Can run detection instead
        clist['tags'] = ut.zipflat(clist['cat'])
        aid_list = ibs.use_images_as_annotations(
            clist['gid'], adjust_percent=0.01, tags_list=clist['tags']
        )
        aid_list

    import wbia.plottool as pt
    from wbia import core_annots

    pt.qt4ensure()
    # annots = ibs.annots()
    # aids = [1, 2]
    # ibs.depc_annot.get('hog', aids , 'hog')
    # ibs.depc_annot.get('chip', aids, 'img')
    for aid in ut.InteractiveIter(ibs.get_valid_aids()):
        hogs = ibs.depc_annot.d.get_hog_hog([aid])
        chips = ibs.depc_annot.d.get_chips_img([aid])
        chip = chips[0]
        hogimg = core_annots.make_hog_block_image(hogs[0])
        pt.clf()
        pt.imshow(hogimg, pnum=(1, 2, 1))
        pt.imshow(chip, pnum=(1, 2, 2))
        fig = pt.gcf()
        fig.show()
        fig.canvas.draw()

    # logger.info(len(groupxs))

    # if False:
    # groupxs = ut.find_duplicate_items(ut.lmap(basename, suffix_list)).values()
    # logger.info(ut.repr3(ut.apply_grouping(all_urls, groupxs)))
    #    # FIX
    #    for fpath, fname in zip(fpath_list, fname_list):
    #        if ut.checkpath(fpath):
    #            ut.move(fpath, join(dirname(fpath), fname))
    #            logger.info('fpath = %r' % (fpath,))

    # import wbia
    # from wbia.dbio import ingest_dataset
    # dbdir = wbia.sysres.lookup_dbdir('WS_ALL')
    # self = ingest_dataset.Ingestable2(dbdir)

    if False:
        # Show overlap matrix
        import wbia.plottool as pt
        import pandas as pd
        import numpy as np

        dict_ = overlaps
        s = pd.Series(dict_, index=pd.MultiIndex.from_tuples(overlaps))
        df = s.unstack()
        lhs, rhs = df.align(df.T)
        df = lhs.add(rhs, fill_value=0).fillna(0)

        label_texts = df.columns.values

        def label_ticks(label_texts):
            import wbia.plottool as pt

            truncated_labels = [repr(lbl[0:100]) for lbl in label_texts]
            ax = pt.gca()
            ax.set_xticks(list(range(len(label_texts))))
            ax.set_xticklabels(truncated_labels)
            [lbl.set_rotation(-55) for lbl in ax.get_xticklabels()]
            [lbl.set_horizontalalignment('left') for lbl in ax.get_xticklabels()]

            # xgrid, ygrid = np.meshgrid(range(len(label_texts)), range(len(label_texts)))
            # pt.plot_surface3d(xgrid, ygrid, disjoint_mat)
            ax.set_yticks(list(range(len(label_texts))))
            ax.set_yticklabels(truncated_labels)
            [lbl.set_horizontalalignment('right') for lbl in ax.get_yticklabels()]
            [lbl.set_verticalalignment('center') for lbl in ax.get_yticklabels()]
            # [lbl.set_rotation(20) for lbl in ax.get_yticklabels()]

        # df = df.sort(axis=0)
        # df = df.sort(axis=1)

        sortx = np.argsort(df.sum(axis=1).values)[::-1]
        df = df.take(sortx, axis=0)
        df = df.take(sortx, axis=1)

        fig = pt.figure(fnum=1)
        fig.clf()
        mat = df.values.astype(np.int32)
        mat[np.diag_indices(len(mat))] = 0
        vmax = mat[(1 - np.eye(len(mat))).astype(np.bool)].max()
        import matplotlib.colors

        norm = matplotlib.colors.Normalize(vmin=0, vmax=vmax, clip=True)
        pt.plt.imshow(mat, cmap='hot', norm=norm, interpolation='none')
        pt.plt.colorbar()
        pt.plt.grid('off')
        label_ticks(label_texts)
        fig.tight_layout()

    # overlap_df = pd.DataFrame.from_dict(overlap_img_list)

    class TmpImage(ut.NiceRepr):
        pass

    from skimage.feature import hog
    from skimage import data, color, exposure
    import wbia.plottool as pt

    image2 = color.rgb2gray(data.astronaut())  # NOQA

    fpath = './GOPR1120.JPG'

    import vtool as vt

    for fpath in [fpath]:
        """
        http://scikit-image.org/docs/dev/auto_examples/plot_hog.html
        """

        image = vt.imread(fpath, grayscale=True)
        image = pt.color_funcs.to_base01(image)

        fig = pt.figure(fnum=2)
        fd, hog_image = hog(
            image,
            orientations=8,
            pixels_per_cell=(16, 16),
            cells_per_block=(1, 1),
            visualise=True,
        )

        fig, (ax1, ax2) = pt.plt.subplots(1, 2, figsize=(8, 4), sharex=True, sharey=True)

        ax1.axis('off')
        ax1.imshow(image, cmap=pt.plt.cm.gray)
        ax1.set_title('Input image')
        ax1.set_adjustable('box-forced')

        # Rescale histogram for better display
        hog_image_rescaled = exposure.rescale_intensity(hog_image, in_range=(0, 0.02))

        ax2.axis('off')
        ax2.imshow(hog_image_rescaled, cmap=pt.plt.cm.gray)
        ax2.set_title('Histogram of Oriented Gradients')
        ax1.set_adjustable('box-forced')
        pt.plt.show()

    # for


def detect_sharks(ibs, gids):
    # import wbia
    # ibs = wbia.opendb('WS_ALL')
    config = {
        'algo': 'yolo',
        'sensitivity': 0.2,
        'config_filepath': ut.truepath(
            '~/work/WS_ALL/localizer_backup/detect.yolo.2.cfg'
        ),
        'weight_filepath': ut.truepath(
            '~/work/WS_ALL/localizer_backup/detect.yolo.2.39000.weights'
        ),
        'class_filepath': ut.truepath(
            '~/work/WS_ALL/localizer_backup/detect.yolo.2.cfg.classes'
        ),
    }
    depc = ibs.depc_image

    # imgsets = ibs.imagesets(text='Injured Sharks')
    # images = ibs.images(imgsets.gids[0])
    images = ibs.images(gids)
    images = images.compress([ext not in ['.gif'] for ext in images.exts])
    gid_list = images.gids

    # result is a tuple:
    # (score, bbox_list, theta_list, conf_list, class_list)
    results_list = depc.get_property('localizations', gid_list, None, config=config)

    results_list2 = []
    multi_gids = []
    failed_gids = []

    # ibs.set_image_imagesettext(failed_gids, ['Fixme'] * len(failed_gids))
    ibs.set_image_imagesettext(multi_gids, ['Fixme2'] * len(multi_gids))

    failed_gids

    for gid, res in zip(gid_list, results_list):
        score, bbox_list, theta_list, conf_list, class_list = res
        if len(bbox_list) == 0:
            failed_gids.append(gid)
        elif len(bbox_list) == 1:
            results_list2.append((gid, bbox_list, theta_list))
        elif len(bbox_list) > 1:
            multi_gids.append(gid)
            idx = conf_list.argmax()
            res2 = (gid, bbox_list[idx : idx + 1], theta_list[idx : idx + 1])
            results_list2.append(res2)

    ut.dict_hist(([t[1].shape[0] for t in results_list]))

    localized_imgs = ibs.images(ut.take_column(results_list2, 0))
    assert all([len(a) == 1 for a in localized_imgs.aids])
    old_annots = ibs.annots(ut.flatten(localized_imgs.aids))
    # old_tags = old_annots.case_tags

    # Override old bboxes
    import numpy as np

    bboxes = np.array(ut.take_column(results_list2, 1))[:, 0, :]
    ibs.set_annot_bboxes(old_annots.aids, bboxes)

    if False:
        import wbia.plottool as pt

        pt.qt4ensure()

        inter = pt.MultiImageInteraction(
            ibs.get_image_paths(ut.take_column(results_list2, 0)),
            bboxes_list=ut.take_column(results_list2, 1),
        )
        inter.dump_to_disk('shark_loc', num=50, prefix='shark_loc')
        inter.start()

        inter = pt.MultiImageInteraction(ibs.get_image_paths(failed_gids))
        inter.start()

        inter = pt.MultiImageInteraction(ibs.get_image_paths(multi_gids))
        inter.start()


def train_part_detector():
    """
    Problem:
        healthy sharks usually have a mostly whole body shot
        injured sharks usually have a close up shot.
        This distribution of images is likely what the injur-shark net is picking up on.

    The goal is to train a detector that looks for things that look
    like the distribution of injured sharks.

    We will run this on healthy sharks to find the parts of
    """
    import wbia

    ibs = wbia.opendb('WS_ALL')
    imgset = ibs.imagesets(text='Injured Sharks')
    injured_annots = imgset.annots[0]  # NOQA

    # config = {
    #    'dim_size': (224, 224),
    #    'resize_dim': 'wh'
    # }

    from pydarknet import Darknet_YOLO_Detector

    data_path = ibs.export_to_xml()
    output_path = join(ibs.get_cachedir(), 'training', 'localizer')
    ut.ensuredir(output_path)
    dark = Darknet_YOLO_Detector()
    results = dark.train(data_path, output_path)
    del dark

    localizer_weight_path, localizer_config_path, localizer_class_path = results
    classifier_model_path = ibs.classifier_train()
    labeler_model_path = ibs.labeler_train()
    output_path = join(ibs.get_cachedir(), 'training', 'detector')
    ut.ensuredir(output_path)
    ut.copy(localizer_weight_path, join(output_path, 'localizer.weights'))
    ut.copy(localizer_config_path, join(output_path, 'localizer.config'))
    ut.copy(localizer_class_path, join(output_path, 'localizer.classes'))
    ut.copy(classifier_model_path, join(output_path, 'classifier.npy'))
    ut.copy(labeler_model_path, join(output_path, 'labeler.npy'))

    # ibs.detector_train()


def purge_ensure_one_annot_per_images(ibs):
    """
    pip install Pipe
    """
    # Purge all but one annotation
    images = ibs.images()
    # images.aids
    groups = images._annot_groups
    import numpy as np

    # Take all but the largest annotations per images
    large_masks = [ut.index_to_boolmask([np.argmax(x)], len(x)) for x in groups.bbox_area]
    small_masks = ut.lmap(ut.not_list, large_masks)
    # Remove all but the largets annotation
    small_aids = ut.zipcompress(groups.aid, small_masks)
    small_aids = ut.flatten(small_aids)

    # Fix any empty images
    images = ibs.images()
    empty_images = ut.where(np.array(images.num_annotations) == 0)
    logger.info('empty_images = %r' % (empty_images,))
    # list(map(basename, map(dirname, images.uris_original)))

    def VecPipe(func):
        import pipe

        @pipe.Pipe
        def wrapped(sequence):
            return map(func, sequence)
            # return (None if item is None else func(item) for item in sequence)

        return wrapped

    name_list = list(images.uris_original | VecPipe(dirname) | VecPipe(basename))
    aids_list = images.aids
    ut.assert_all_eq(list(aids_list | VecPipe(len)))
    annots = ibs.annots(ut.flatten(aids_list))
    annots.names = name_list


def shark_misc():
    import wbia

    ibs = wbia.opendb('WS_ALL')
    aid_list = ibs.get_valid_aids()
    flag_list = ibs.get_annot_been_adjusted(aid_list)
    adjusted_aids = ut.compress(aid_list, flag_list)
    return adjusted_aids

    # if False:
    #    # TRY TO FIGURE OUT WHY URLS ARE MISSING IN STEP 1
    #    encounter_to_parsed1 = parsed1.group_items('encounter')
    #    encounter_to_parsed2 = parsed2.group_items('encounter')

    #    url_to_parsed1 = parsed1.group_items('img_url')
    #    url_to_parsed2 = parsed2.group_items('img_url')

    #    def set_overlap(set1, set2):
    #        set1 = set(set1)
    #        set2 = set(set2)
    #        return ut.odict([
    #            ('s1', len(set1)),
    #            ('s2', len(set2)),
    #            ('isect', len(set1.intersection(set2))),
    #            ('union', len(set1.union(set2))),
    #            ('s1 - s2', len(set1.difference(set2))),
    #            ('s2 - s1', len(set2.difference(set1))),
    #        ])
    #    logger.info('encounter overlap: ' + ut.repr3(set_overlap(encounter_to_parsed1, encounter_to_parsed2)))
    #    logger.info('url overlap: ' + ut.repr3(set_overlap(url_to_parsed1, url_to_parsed2)))

    #    url1 = list(url_to_parsed1.keys())
    #    url2 = list(url_to_parsed2.keys())
    #    # remove common prefixes
    #    from os.path import commonprefix, basename  # NOQA
    #    cp1 = commonprefix(url1)
    #    cp2 = commonprefix(url2)
    #    #suffix1 = sorted([u[len(cp1):].lower() for u in url1])
    #    #suffix2 = sorted([u[len(cp2):].lower() for u in url2])
    #    suffix1 = sorted([u[len(cp1):] for u in url1])
    #    suffix2 = sorted([u[len(cp2):] for u in url2])
    #    logger.info('suffix overlap: ' + ut.repr3(set_overlap(suffix1, suffix2)))
    #    set1 = set(suffix1)
    #    set2 = set(suffix2)
    #    only1 = list(set1 - set1.intersection(set2))
    #    only2 = list(set2 - set1.intersection(set2))

    #    import numpy as np
    #    for suf in ut.ProgIter(only2, bs=True):
    #        dist = np.array(ut.edit_distance(suf, only1))
    #        idx = ut.argsort(dist)[0:3]
    #        if dist[idx][0] < 3:
    #            close = ut.take(only1, idx)
    #            logger.info('---')
    #            logger.info('suf = %r' % (join(cp2, suf),))
    #            logger.info('close = %s' % (ut.repr3([join(cp1, c) for c in close]),))
    #            logger.info('---')
    #            break

    #    # Associate keywords with original images
    #    #lower_urls = [x.lower() for x in parsed['img_url']]
    #    url_to_idx = ut.make_index_lookup(parsed1['img_url'])
    #    parsed1['keywords'] = [[] for _ in range(len(parsed1))]
    #    for url, keys in url_to_keys.items():
    #        # hack because urls are note in the same format
    #        url = url.replace('wildbook_data_dir', 'shepherd_data_dir')
    #        url = url.lower()
    #        if url in url_to_idx:
    #            idx = url_to_idx[url]
    #            parsed1['keywords'][idx].extend(keys)

    # healthy_annots = ibs.annots(ibs.imagesets(text='Non-Injured Sharks').aids[0])
    # ibs.set_annot_prop('healthy', healthy_annots.aids, [True] * len(healthy_annots))
    # ['healthy' in t and len(t) > 0 for t in single_annots.case_tags]
    # healthy_tags = []

    # ut.find_duplicate_items(cur_img_uuids)
    # ut.find_duplicate_items(new_img_uuids)
    # cur_uuids = set(cur_img_uuids)
    # new_uuids = set(new_img_uuids)
    # both_uuids = new_uuids.intersection(cur_uuids)
    # only_cur = cur_uuids - both_uuids
    # only_new = new_uuids - both_uuids
    # logger.info('len(cur_uuids) = %r' % (len(cur_uuids)))
    # logger.info('len(new_uuids) = %r' % (len(new_uuids)))
    # logger.info('len(both_uuids) = %r' % (len(both_uuids)))
    # logger.info('len(only_cur) = %r' % (len(only_cur)))
    # logger.info('len(only_new) = %r' % (len(only_new)))

    # Ensure that data in both sets are syncronized
    # images_both = []

    # if False:
    #    logger.info('Removing small images')
    #    import numpy as np
    #    import vtool as vt
    #    imgsize_list = np.array([vt.open_image_size(gpath) for gpath in parsed['new_fpath']])
    #    sqrt_area_list = np.sqrt(np.prod(imgsize_list, axis=1))
    #    areq_flags_list = sqrt_area_list >= 750
    #    parsed = parsed.compress(areq_flags_list)



# -*- coding: utf-8 -*-
import logging
import re
import functools
import operator as op
import utool as ut
import numpy as np
import copy
import six

(print, rrr, profile) = ut.inject2(__name__, '[depbase]')
logger = logging.getLogger('wbia')


class StackedConfig(ut.DictLike, ut.HashComparable):
    """
    Manages a list of configurations
    """

    def __init__(self, config_list):
        self._orig_config_list = config_list
        # Cast all inputs to config classes
        self._new_config_list = [
            cfg if hasattr(cfg, 'get_cfgstr') else make_configclass(cfg, '')
            for cfg in self._orig_config_list
        ]
        # Parse out items
        self._items = ut.flatten(
            [
                list(cfg.parse_items())
                if hasattr(cfg, 'parse_items')
                else list(cfg.items())
                for cfg in self._orig_config_list
            ]
        )
        for key, val in self._items:
            setattr(self, key, val)
        # self.keys = ut.flatten(list(cfg.keys()) for cfg in self.config_list)

    def get_cfgstr(self):
        cfgstr_list = [cfg.get_cfgstr() for cfg in self._new_config_list]
        cfgstr = '_'.join(cfgstr_list)
        return cfgstr

    def keys(self):
        return ut.take_column(self._items, 0)

    def __hash__(cfg):
        """ Needed for comparison operators """
        return hash(cfg.get_cfgstr())

    def getitem(self, key):
        try:
            return getattr(self, key)
        except AttributeError as ex:
            raise KeyError(ex)


# @six.add_metaclass(ut.HashComparableMetaclass)
@functools.total_ordering
class Config(ut.NiceRepr, ut.DictLike):
    r"""
    Base class for heirarchical config
    need to overwrite get_param_info_list

    CommandLine:
        python -m dtool.base Config

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.dtool.base import *  # NOQA
        >>> cfg1 = Config.from_dict({'a': 1, 'b': 2})
        >>> cfg2 = Config.from_dict({'a': 2, 'b': 2})
        >>> # Must be hashable and orderable
        >>> hash(cfg1)
        >>> cfg1 > cfg2

    """

    def __init__(cfg, **kwargs):
        cfg._parent = None
        cfg.initialize_params(**kwargs)

    def deepcopy(cfg):
        cfg2 = copy.deepcopy(cfg)
        cfg2._subconfig_attrs = copy.deepcopy(cfg._subconfig_attrs)
        cfg2._subconfig_names = copy.deepcopy(cfg._subconfig_names)
        try:
            cfg2._param_info_list = copy.deepcopy(cfg._param_info_list)
        except AttributeError:
            pass
        return cfg2

    def __nice__(cfg):
        return cfg.get_cfgstr(with_name=False)

    def __lt__(self, other):
        """ hash comparable broke in python3 """
        return ut.compare_instance(op.lt, self, other)

    def __eq__(self, other):
        """ hash comparable broke in python3 """
        return ut.compare_instance(op.eq, self, other)

    def __hash__(cfg):
        """ Needed for comparison operators """
        return hash(cfg.get_cfgstr())

    def get_config_name(cfg, **kwargs):
        """ the user might want to overwrite this function """
        # VERY HACKY
        config_name = cfg.__class__.__name__.replace('Config', '')
        config_name = re.sub('_$', '', config_name)
        return config_name

    def get_varnames(cfg):
        return [pi.varname for pi in cfg.get_param_info_list()] + cfg._subconfig_attrs

    def update(cfg, **kwargs):
        """
        Overwrites default DictLike update for only keys that exist.
        Non-existing key are ignored.

        Note:
            prefixed keys in the form <classname>_<key> will be just be
            interpreted as <key>

        CommandLine:
            python -m dtool.base update --show

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.base import *  # NOQA
            >>> from wbia.dtool.example_depcache import DummyVsManyConfig
            >>> cfg = DummyVsManyConfig()
            >>> cfg.update(DummyAlgo_version=4)
            >>> print(cfg)
        """
        # FIXME: currently can't update subconfigs based on namespaces
        # and non-namespaced vars are in the context of the root level.
        # self_keys = set(cfg.__dict__.keys())
        # self_keys.append(cfg.get_varnames())
        _aliases = cfg._make_key_alias_checker()
        self_keys = set(cfg.keys())
        for key, val in six.iteritems(kwargs):
            # update only existing keys or namespace prefixed keys
            for k in _aliases(key):
                if k in self_keys:
                    cfg.setitem(k, val)
                    break

    def pop_update(cfg, other):
        """
        Updates based on other, while popping off used arguments.
        (useful for testing if a parameter was unused or misspelled)

        Doctest:
            >>> from wbia.dtool.base import *  # NOQA
            >>> from wbia import dtool as dt
            >>> cfg = dt.Config.from_dict({'a': 1, 'b': 2, 'c': 3})
            >>> other = {'a': 5, 'e': 2}
            >>> cfg.pop_update(other)
            >>> assert cfg['a'] == 5
            >>> assert len(other) == 1 and 'a' not in other
        """
        _aliases = cfg._make_key_alias_checker()
        self_keys = set(cfg.keys())
        for key in list(other.keys()):
            # update only existing keys or namespace prefixed keys
            for k in _aliases(key):
                if k in self_keys:
                    val = other.pop(key)
                    cfg.setitem(k, val)
                    break

    def _make_key_alias_checker(cfg):
        prefixes = (cfg.get_config_name(), cfg.__class__.__name__)

        def _aliases(key):
            yield key
            for part in prefixes:
                prefix = part + '_'
                if key.startswith(prefix):
                    key_alias = key[len(prefix) :]
                    yield key_alias

        return _aliases

    def update2(cfg, *args, **kwargs):
        """
        Overwrites default DictLike update for only keys that exist.
        Non-existing key are ignored.
        Also updates nested configs.

        Note:
            prefixed keys in the form <classname>_<key> will be just be
            interpreted as <key>

        CommandLine:
            python -m dtool.base update --show

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.base import *  # NOQA
            >>> from wbia import dtool as dt
            >>> cfg = dt.Config.from_dict({
            >>>     'a': 1,
            >>>     'b': 2,
            >>>     'c': 3,
            >>>     'sub1': dt.Config.from_dict({
            >>>         'x': 'x',
            >>>         'y': {'z', 'x'},
            >>>         'c': 33,
            >>>     }),
            >>>     'sub2': dt.Config.from_dict({
            >>>         's': [1, 2, 3],
            >>>         't': (1, 2, 3),
            >>>         'c': 42,
            >>>         'sub3': dt.Config.from_dict({
            >>>             'b': 99,
            >>>             'c': 88,
            >>>         }),
            >>>     }),
            >>> })
            >>> kwargs = {'c': 10}
            >>> cfg.update2(c=10, y={1,2})
            >>> assert cfg.c == 10
            >>> assert cfg.sub1.c == 10
            >>> assert cfg.sub2.c == 10
            >>> assert cfg.sub2.sub3.c == 10
            >>> assert cfg.sub1.y == {1, 2}
        """
        if len(args) > 1:
            raise ValueError('only specify one arg')
        elif len(args) == 1:
            kwargs.update(args[0])
        return list(cfg._update2(kwargs))

    def _update2(cfg, kwargs):
        # yields a list of keys updated as they happen
        _aliases = cfg._make_key_alias_checker()
        for key, val in cfg.native_items():
            for k in _aliases(key):
                if k in kwargs:
                    cfg.setitem(k, kwargs[k])
                    yield k
                    break
        for key, val in cfg.nested_items():
            val = cfg[key]
            if isinstance(val, Config):
                for k in val._update2(kwargs):
                    yield k

    def nested_items(cfg):
        for key in cfg.keys():
            val = cfg[key]
            if isinstance(val, Config):
                yield key, val

    def native_items(cfg):
        for key in cfg.keys():
            val = cfg[key]
            if not isinstance(val, Config):
                yield key, val

    def initialize_params(cfg, **kwargs):
        """ Initializes config class attributes based on params info list """
        # logger.info("INIT PARAMS")
        for pi in cfg.get_param_info_list():
            setattr(cfg, pi.varname, pi.default)

        # SO HACKY
        # Hacks in implicit edges from nodes to the algorithm
        # using their subconfigurations
        cfg._subconfig_attrs = []
        cfg._subconfig_names = []
        _sub_config_list = cfg.get_sub_config_list()
        if _sub_config_list:
            for subclass in _sub_config_list:
                # subclass.static_config_name()
                subcfg = subclass()
                subcfg_name = subcfg.get_config_name()
                subcfg_attr = ut.to_underscore_case(subcfg_name) + '_cfg'
                setattr(cfg, subcfg_attr, subcfg)
                cfg._subconfig_names.append(subcfg_name)
                cfg._subconfig_attrs.append(subcfg_attr)
                subcfg.update(**kwargs)
        cfg.update(**kwargs)

    def get_sub_config_list(cfg):
        if hasattr(cfg, '_sub_config_list'):
            return cfg._sub_config_list
        else:
            return []

    def parse_namespace_config_items(cfg):
        """
        Recursively extracts key, val pairs from Config objects
        into a flat list. (there must not be name conflicts)
        """
        param_list = []
        seen = set([])
        for item in cfg.items():
            key, val = item
            if hasattr(val, 'parse_namespace_config_items'):
                child_cfg = val
                child_params = child_cfg.parse_namespace_config_items()
                param_list.extend(child_params)
            elif hasattr(val, 'parse_items'):
                # hack for ut.Pref configs
                name = val.get_config_name()
                for key, val in val.parse_items():
                    if key in seen:
                        logger.info(
                            '[Config] WARNING: key=%r appears more than once' % (key,)
                        )
                    seen.add(key)
                    # Incorporate namespace
                    param_list.append((name, key, val))
            elif key.startswith('_'):
                pass
            else:
                if key in seen:
                    logger.info(
                        '[Config] WARNING: key=%r appears more than once' % (key,)
                    )
                seen.add(key)
                # Incorporate namespace
                name = cfg.get_config_name()
                param_list.append((name, key, val))
        return param_list

    def parse_items(cfg):
        r"""
        Returns:
            list: param_list

        CommandLine:
            python -m dtool.base --exec-parse_items

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.base import *  # NOQA
            >>> from wbia.dtool.example_depcache import DummyVsManyConfig
            >>> cfg = DummyVsManyConfig()
            >>> param_list = cfg.parse_items()
            >>> result = ('param_list = %s' % (ut.repr2(param_list, nl=1),))
            >>> print(result)
        """
        namespace_param_list = cfg.parse_namespace_config_items()
        param_names = ut.get_list_column(namespace_param_list, 1)
        needs_namespace_keys = ut.find_duplicate_items(param_names)
        param_list = ut.get_list_column(namespace_param_list, [1, 2])
        # prepend namespaces to variables that need it
        for idx in ut.flatten(needs_namespace_keys.values()):
            name = namespace_param_list[idx][0]
            param_list[idx][0] = name + '_' + param_list[idx][0]
        duplicate_keys = ut.find_duplicate_items(ut.get_list_column(param_list, 0))
        # hack to let version through
        # import utool
        # with utool.embed_on_exception_context:
        assert len(duplicate_keys) == 0, (
            'Configs have duplicate names: %r' % duplicate_keys
        )
        return param_list

    def get_cfgstr_list(cfg, ignore_keys=None, with_name=True, **kwargs):
        """ default get_cfgstr_list, can be overrided by a config object """
        if ignore_keys is not None:
            itemstr_list = [
                pi.get_itemstr(cfg)
                for pi in cfg.get_param_info_list()
                if pi.varname not in ignore_keys
            ]
        else:
            itemstr_list = [pi.get_itemstr(cfg) for pi in cfg.get_param_info_list()]
        filtered_itemstr_list = list(filter(len, itemstr_list))
        if with_name:
            config_name = cfg.get_config_name()
        else:
            config_name = ''
        body = ','.join(filtered_itemstr_list)
        cfgstr = ''.join([config_name, '(', body, ')'])
        return cfgstr

    def get_cfgstr(cfg, **kwargs):
        str_ = ''.join(cfg.get_cfgstr_list(**kwargs))
        return '_'.join(
            [str_]
            + [cfg[subcfg_attr].get_cfgstr() for subcfg_attr in cfg._subconfig_attrs]
        )

    def get_param_info_dict(cfg):
        param_info_list = cfg.get_param_info_list()
        param_info_dict = {pi.varname: pi for pi in param_info_list}
        return param_info_dict

    def assert_self_types(cfg, verbose=True):
        if verbose:
            logger.info('Assert self types of cfg=%r' % (cfg,))
        pi_dict = cfg.get_param_info_dict()
        for key in cfg.keys():
            pi = pi_dict[key]
            value = cfg[key]
            pi.error_if_invalid_value(value)
        if verbose:
            logger.info('... checks passed')

    def getinfo(cfg, key):
        pass

    def get_hashid(cfg):
        return ut.hashstr27(cfg.get_cfgstr())

    def keys(cfg):
        """ Required for DictLike interface """
        return cfg.get_varnames()

    def getitem(cfg, key):
        """ Required for DictLike interface """
        try:
            return getattr(cfg, key)
        except AttributeError as ex:
            raise KeyError(ex)

    def get(qparams, key, *d):
        """ get a paramater value by string """
        ERROR_ON_DEFAULT = False
        if ERROR_ON_DEFAULT:
            return getattr(qparams, key)
        else:
            return getattr(qparams, key, *d)

    def setitem(cfg, key, value):
        """ Required for DictLike interface """
        # TODO; check for valid config setting
        pi_dict = cfg.get_param_info_dict()
        pi = pi_dict[key]
        pi.error_if_invalid_value(value)
        return setattr(cfg, key, value)

    def get_param_info_list(cfg):
        try:
            return cfg._param_info_list
        except AttributeError:
            raise NotImplementedError(
                'Need to define _param_info_list or get_param_info_list'
            )

    @classmethod
    def from_argv_dict(cls, **kwargs):
        """
        handy command line tool
        ut.parse_argv_cfg
        """
        cfg = cls(**kwargs)
        new_vals = ut.parse_dict_from_argv(cfg)
        cfg.update(**new_vals)
        return cfg

    @classmethod
    def from_argv_cfgs(cls):
        """
        handy command line tool
        """
        cfg = cls()
        name = cfg.get_config_name()
        # name = cls.static_config_name()
        argname = '--' + name
        if hasattr(cfg, '_alias'):
            argname = (argname, '--' + cfg._alias)
        # if hasattr(cls, '_alias'):
        #    argname = (argname, '--' + cls._alias)
        new_vals_list = ut.parse_argv_cfg(argname)
        self_list = [cls(**new_vals) for new_vals in new_vals_list]
        return self_list

    @classmethod
    def from_dict(cls, dict_, tablename=None):
        r"""
        Args:
            dict_ (dict_):  a dictionary
            tablename (None): (default = None)

        Returns:
            list: param_info_list

        CommandLine:
            python -m dtool.base Config.from_dict --show

        Example:
            >>> # DISABLE_DOCTEST
            >>> from wbia.dtool.base import *  # NOQA
            >>> cls = Config
            >>> dict_ = {'K': 1, 'Knorm': 5, 'min_pername': 1, 'max_pername': 1,}
            >>> tablename = None
            >>> config = cls.from_dict(dict_, tablename)
            >>> print(config)
            >>> # xdoctest: +REQUIRES(--show)
            >>> ut.quit_if_noshow()
            >>> dlg = config.make_qt_dialog(
            >>>     title='Confirm Merge Query',
            >>>     msg='Confirm')
            >>> dlg.resize(700, 500)
            >>> dlg.show()
            >>> import wbia.plottool as pt
            >>> self = dlg.widget
            >>> guitool.qtapp_loop(qwin=dlg)
            >>> updated_config = self.config  # NOQA
            >>> print('updated_config = %r' % (updated_config,))
        """
        UnnamedConfig = cls.class_from_dict(dict_, tablename)
        config = UnnamedConfig()
        return config

    @classmethod
    def class_from_dict(cls, dict_, tablename=None):
        if tablename is None:
            tablename = 'Unnamed'
        UnnamedConfig = make_configclass(dict_, tablename)
        return UnnamedConfig

    def make_qt_dialog(cfg, parent=None, title='Edit Config', msg='Confim'):
        import wbia.guitool as gt

        gt.ensure_qapp()  # must be ensured before any embeding
        dlg = gt.ConfigConfirmWidget.as_dialog(title=title, msg=msg, config=cfg)
        dlg.resize(700, 500)
        dlg.show()
        return dlg

    def getstate_todict_recursive(cfg):
        from wbia import dtool

        _dict = cfg.asdict()
        _dict2 = {}
        for key, val in _dict.items():
            if isinstance(val, dtool.Config):
                # val = val.asdict()
                try:
                    val = val.getstate_todict_recursive()
                except Exception:
                    val = getstate_todict_recursive(val)  # NOQA
                _dict2[key] = val
            else:
                _dict2[key] = val
        return _dict2

    def __getstate__(cfg):
        """
        FIXME

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.base import *  # NOQA
            >>> from wbia.dtool.example_depcache import DummyKptsConfig
            >>> from six.moves import cPickle as pickle
            >>> cfg = DummyKptsConfig()
            >>> ser = pickle.dumps(cfg)
            >>> cfg2 = pickle.loads(ser)
            >>> assert cfg == cfg2
            >>> assert cfg is not cfg2

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.base import *  # NOQA
            >>> from wbia.dtool.example_depcache import DummyVsManyConfig
            >>> from six.moves import cPickle as pickle
            >>> cfg = DummyVsManyConfig()
            >>> state = cfg.__getstate__()
            >>> cfg2 = DummyVsManyConfig()
            >>> serialized = pickle.dumps(cfg)
            >>> unserialized = pickle.loads(serialized)
            >>> assert cfg == unserialized
            >>> assert cfg is not unserialized
        """
        # from wbia import dtool
        # _dict = cfg.asdict()
        # _dict2 = {}
        # for key, val in _dict.items():
        #    if isinstance(val, dtool.Config):
        #        val = val.asdict()
        #    _dict2[key] = val
        # return {'dtool.Config': _dict2}
        return cfg.__dict__

    def __setstate__(cfg, state):
        cfg.__dict__.update(**state)
        # cfg.initialize_params()
        # cfg.update(**state)

    # @classmethod
    # def static_config_name(cls):
    #    class_str = str(cls)
    #    full_class_str = class_str.replace('<class \'', '').replace('\'>', '')
    #    config_name = splitext(full_class_str)[1][1:].replace('Config', '')
    #    return config_name


def make_configclass(dict_, tablename):
    """ Creates a custom config class from a dict """

    def rectify_item(key, val):
        if val is None:
            return ut.ParamInfo(key, val)
        elif isinstance(val, ut.ParamInfo):
            if val.varname is None:
                # Copy and assign a new varname
                pi = copy.deepcopy(val)
                pi.varname = key
            else:
                pi = val
                assert pi.varname == key, 'Given varname=%r does not match key=%r' % (
                    pi.varname,
                    key,
                )
            return pi
        else:
            if isinstance(val, Config):
                # Set table name from key when doing nested from dicts
                if val.__class__.__name__ == 'UnnamedConfig':
                    val.__class__.__name__ = str(key + 'Config')
            return ut.ParamInfo(key, val, type_=type(val))

    param_info_list = [rectify_item(key, val) for key, val in dict_.items()]
    return from_param_info_list(param_info_list, tablename)


def from_param_info_list(param_info_list, tablename='Unnamed'):
    from wbia import dtool

    class UnnamedConfig(dtool.Config):
        _param_info_list = param_info_list

    UnnamedConfig.__name__ = str(tablename + 'Config')
    return UnnamedConfig


class IBEISRequestHacks(object):
    _isnewreq = True

    @property
    def ibs(request):
        """ HACK specific to wbia """
        if request.depc is None:
            return None
        return request.depc.controller

    @property
    def qannots(self):
        return self.ibs.annots(self.qaids, self.params)

    @property
    def dannots(self):
        return self.ibs.annots(self.daids, self.params)

    def get_qreq_annot_nids(self, aids):
        # VERY HACKY. To be just hacky it should store
        # the nids as a state, but whatever...
        # devleopment time constraints and whatnot
        return self.ibs.get_annot_nids(aids)
        # return self.ibs.annots(self.daids, self.params)

    @property
    def extern_query_config2(request):
        return request.params

    @property
    def extern_data_config2(request):
        return request.params

    #

    # def get_external_data_config2(request):
    #    # HACK
    #    #return None
    #    #logger.info('[d] request.params = %r' % (request.params,))
    #    return request.params

    # def get_external_query_config2(request):
    #    # HACK
    #    #return None
    #    #logger.info('[q] request.params = %r' % (request.params,))
    #    return request.params


def config_graph_subattrs(cfg, depc):
    # TODO: if this hack is fully completed need a way of getting the
    # full config belonging to both chip + feat
    # cfg = request.config.feat_cfg
    import networkx as netx

    tablename = ut.invert_dict(depc.configclass_dict)[cfg.__class__]
    # tablename = cfg.get_config_name()
    ancestors = netx.dag.ancestors(depc.graph, tablename)
    subconfigs_ = ut.dict_take(depc.configclass_dict, ancestors, None)
    subconfigs = ut.filter_Nones(subconfigs_)  # NOQA


@ut.reloadable_class
class BaseRequest(IBEISRequestHacks, ut.NiceRepr):
    r"""
    Class that maintains both an algorithm, inputs, and a config.
    """

    @staticmethod
    def static_new(cls, depc, parent_rowids, cfgdict=None, tablename=None):
        """ hack for autoreload """
        request = cls()
        if tablename is None:
            try:
                if hasattr(cls, '_tablename'):
                    tablename = cls._tablename
                else:
                    tablename = ut.invert_dict(depc.requestclass_dict)[cls]
            except Exception as ex:
                ut.printex(ex, 'tablename must be given')
                raise
        request.tablename = tablename
        request.parent_rowids = parent_rowids
        request.depc = depc
        if cfgdict is None:
            cfgdict = {}
        configclass = depc.configclass_dict[tablename]
        config = configclass(**cfgdict)
        request.config = config
        # HACK FOR IBEIS
        request.params = dict(config.parse_items())
        # HACK-ier FOR BACKWARDS COMPATABILITY
        if True:
            # params.featweight_cfgstr = query_cfg._featweight_cfg.get_cfgstr()
            # TODO: if this hack is fully completed need a way of getting the
            # full config belonging to both chip + feat
            try:
                request.params['chip_cfgstr'] = config.chip_cfg.get_cfgstr()
                request.params['chip_cfg_dict'] = config.chip_cfg.asdict()
                request.params['feat_cfgstr'] = config.feat_cfg.get_cfgstr()
                request.params['hesaff_params'] = config.feat_cfg.get_hesaff_params()
                request.params['featweight_cfgstr'] = config.feat_weight_cfg.get_cfgstr()
            except AttributeError:
                pass
        request.qparams = ut.DynStruct()
        for key, val in request.params.items():
            setattr(request.qparams, key, val)

        return request

    @classmethod
    def new(cls, depc, parent_rowids, cfgdict=None, tablename=None):
        return cls.static_new(cls, depc, parent_rowids, cfgdict, tablename)

    def _get_rootset_hashid(request, root_rowids, prefix):
        uuid_type = 'V'
        label = ''.join((prefix, uuid_type, 'UUIDS'))
        # Hack: allow general specification of uuid types
        uuid_list = request.depc.get_root_uuid(root_rowids)
        # uuid_hashid = ut.hashstr_arr27(uuid_list, label, pathsafe=True)
        uuid_hashid = ut.hashstr_arr27(uuid_list, label, pathsafe=False)
        # TODO: uuid_hashid = ut.hashid_arr(uuid_list, label=label)
        return uuid_hashid

    def get_cfgstr(request, with_input=False, with_pipe=True, **kwargs):
        r"""
        main cfgstring used to identify the 'querytype'
        """
        cfgstr_list = []
        if with_input:
            cfgstr_list.append(request.get_input_hashid())
        if with_pipe:
            cfgstr_list.append(request.get_pipe_cfgstr())
        cfgstr = '_'.join(cfgstr_list)
        return cfgstr

    def get_input_hashid(request):
        raise NotImplementedError('abstract class method')

    def get_pipe_cfgstr(request):
        return request.config.get_cfgstr()

    def get_pipe_hashid(request):
        return ut.hashstr27(request.get_pipe_cfgstr())

    def ensure_dependencies(request):
        r"""
        CommandLine:
            python -m dtool.base --exec-BaseRequest.ensure_dependencies

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.base import *  # NOQA
            >>> from wbia.dtool.example_depcache import testdata_depc
            >>> depc = testdata_depc()
            >>> request = depc.new_request('vsmany', [1, 2], [2, 3, 4])
            >>> request.ensure_dependencies()
        """
        import networkx as nx

        depc = request.depc
        if False:
            dependencies = nx.ancestors(depc.graph, request.tablename)
            subgraph = depc.graph.subgraph(set.union(dependencies, {request.tablename}))
            dependency_order = nx.topological_sort(subgraph)
            root = dependency_order[0]
            [
                nx.algorithms.dijkstra_path(subgraph, root, start)[:-1]
                + nx.algorithms.dijkstra_path(subgraph, start, request.tablename)
                for start in dependency_order
            ]
        graph = depc.graph
        root = list(nx.topological_sort(graph))[0]
        edges = graph.edges()
        # parent_to_children = ut.edges_to_adjacency_list(edges)
        child_to_parents = ut.edges_to_adjacency_list([t[::-1] for t in edges])
        to_root = {
            request.tablename: ut.paths_to_root(request.tablename, root, child_to_parents)
        }
        from_root = ut.reverse_path(to_root, root, child_to_parents)
        dependency_levels_ = ut.get_levels(from_root)
        dependency_levels = ut.longest_levels(dependency_levels_)

        true_order = ut.flatten(dependency_levels)[1:-1]
        # logger.info('[req] Ensuring %s request dependencies: %r' % (request, true_order,))
        ut.colorprint(
            '[req] Ensuring request %s dependencies: %r' % (request, true_order),
            'yellow',
        )
        for tablename in true_order:
            table = depc[tablename]
            if table.ismulti:
                pass
            else:
                # HACK FOR IBEIS
                all_aids = ut.flat_unique(request.qaids, request.daids)
                depc.get_rowids(tablename, all_aids)
                pass
            pass

        # zip(depc.get_implicit_edges())
        # zip(depc.get_implicit_edges())

        # raise NotImplementedError('todo')
        # depc = request.depc
        # parent_rowids = request.parent_rowids
        # config = request.config
        # rowid_dict = depc.get_all_descendant_rowids(
        #    request.tablename, root_rowids, config=config)
        pass

    def execute(request, parent_rowids=None, use_cache=None, postprocess=True):
        ut.colorprint('[req] Executing request %s' % (request,), 'yellow')
        table = request.depc[request.tablename]
        if use_cache is None:
            use_cache = not ut.get_argflag('--nocache')
        if parent_rowids is None:
            parent_rowids = request.parent_rowids
        # Compute and cache any uncomputed results
        rowids = table.get_rowid(parent_rowids, config=request, recompute=not use_cache)
        # Load all results
        result_list = table.get_row_data(rowids)
        if postprocess and hasattr(request, 'postprocess_execute'):
            logger.info('Converting results')
            result_list = request.postprocess_execute(parent_rowids, result_list)
            pass
        return result_list

    def __getstate__(request):
        state_dict = request.__dict__.copy()
        # SUPER HACK
        state_dict['dbdir'] = request.depc.controller.get_dbdir()
        del state_dict['depc']
        del state_dict['config']
        return state_dict

    def __setstate__(request, state_dict):
        import wbia

        dbdir = state_dict['dbdir']
        del state_dict['dbdir']
        params = state_dict['params']
        depc = wbia.opendb(dbdir=dbdir, web=False).depc
        configclass = depc.configclass_dict[state_dict['tablename']]
        config = configclass(**params)
        state_dict['depc'] = depc
        state_dict['config'] = config
        request.__dict__.update(state_dict)


class AnnotSimiliarity(object):
    def get_query_hashid(request):
        return request._get_rootset_hashid(request.qaids, 'Q')

    def get_data_hashid(request):
        return request._get_rootset_hashid(request.daids, 'D')


@ut.reloadable_class
class VsOneSimilarityRequest(BaseRequest, AnnotSimiliarity):
    r"""
    Similarity request for pairwise scores

    References:
        https://thingspython.wordpress.com/2010/09/27/
        another-super-wrinkle-raising-typeerror/

    CommandLine:
        python -m dtool.base --exec-VsOneSimilarityRequest

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.dtool.base import *  # NOQA
        >>> from wbia.dtool.example_depcache import testdata_depc
        >>> qaid_list = [1, 2, 3, 5]
        >>> daid_list = [2, 3, 4]
        >>> depc = testdata_depc()
        >>> request = depc.new_request('vsone', qaid_list, daid_list)
        >>> results = request.execute()
        >>> # Test that adding a query / data id only recomputes necessary items
        >>> request2 = depc.new_request('vsone', qaid_list + [4], daid_list + [5])
        >>> results2 = request2.execute()
        >>> print('results = %r' % (results,))
        >>> print('results2 = %r' % (results2,))
        >>> ut.assert_eq(len(results), 10, 'incorrect num output')
        >>> ut.assert_eq(len(results2), 16, 'incorrect num output')
    """

    _symmetric = False

    @classmethod
    def new(cls, depc, qaid_list, daid_list, cfgdict=None, tablename=None):
        parent_rowids = cls.make_parent_rowids(qaid_list, daid_list)
        parent_rowids = list(ut.product_nonsame(qaid_list, daid_list))
        request = cls.static_new(cls, depc, parent_rowids, cfgdict, tablename)
        request.qaids = safeop(np.array, qaid_list)
        request.daids = safeop(np.array, daid_list)
        return request

    @staticmethod
    def make_parent_rowids(qaid_list, daid_list):
        return list(ut.product_nonsame(qaid_list, daid_list))

    @property
    def parent_rowids_T(request):
        return ut.list_transpose(request.parent_rowids)

    def execute(request, parent_rowids=None, use_cache=None, postprocess=True):
        """ HACKY REIMPLEMENTATION """
        ut.colorprint('[req] Executing request %s' % (request,), 'yellow')
        table = request.depc[request.tablename]
        if use_cache is None:
            use_cache = not ut.get_argflag('--nocache')
        if parent_rowids is None:
            parent_rowids = request.parent_rowids
        else:
            # previously defined in execute subset
            # subparent_rowids = request.make_parent_rowids(
            # qaids, request.daids)
            logger.info('given %d specific parent_rowids' % (len(parent_rowids),))

        # vsone hack (i,j) same as (j,i)
        if request._symmetric:
            import vtool as vt

            directed_edges = np.array(parent_rowids)
            undirected_edges = vt.to_undirected_edges(directed_edges)
            edge_ids = vt.compute_unique_data_ids(undirected_edges)
            unique_rows, unique_rowx, inverse_idx = np.unique(
                edge_ids, return_index=True, return_inverse=True
            )
            parent_rowids_ = ut.take(parent_rowids, unique_rowx)
        else:
            parent_rowids_ = parent_rowids

        # Compute and cache any uncomputed results
        rowids = table.get_rowid(parent_rowids_, config=request, recompute=not use_cache)
        # Load all results
        result_list = table.get_row_data(rowids)

        if request._symmetric:
            result_list = ut.take(result_list, inverse_idx)

        if postprocess and hasattr(request, 'postprocess_execute'):
            logger.info('Converting results')
            result_list = request.postprocess_execute(parent_rowids, result_list)
            pass
        return result_list

    def get_input_hashid(request):
        return '_'.join([request.get_query_hashid(), request.get_data_hashid()])

    def __nice__(request):
        dbname = (
            None
            if request.depc is None or request.depc.controller is None
            else request.depc.controller.get_dbname()
        )
        infostr_ = 'nQ=%s, nD=%s, nP=%d %s' % (
            len(request.qaids),
            len(request.daids),
            len(request.parent_rowids),
            request.get_pipe_hashid(),
        )
        return '(%s) %s' % (dbname, infostr_)


@ut.reloadable_class
class VsManySimilarityRequest(BaseRequest, AnnotSimiliarity):
    r"""
    Request for one-vs-many simlarity

    CommandLine:
        python -m dtool.base --exec-VsManySimilarityRequest

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.dtool.base import *  # NOQA
        >>> from wbia.dtool.example_depcache import testdata_depc
        >>> qaid_list = [1, 2]
        >>> daid_list = [2, 3, 4]
        >>> depc = testdata_depc()
        >>> request = depc.new_request('vsmany', qaid_list, daid_list)
        >>> request.ensure_dependencies()
        >>> results = request.execute()
        >>> # Test dependence on data
        >>> request2 = depc.new_request('vsmany', qaid_list + [3], daid_list + [5])
        >>> results2 = request2.execute()
        >>> print('results = %r' % (results,))
        >>> print('results2 = %r' % (results2,))
        >>> assert len(results) == 2, 'incorrect num output'
        >>> assert len(results2) == 3, 'incorrect num output'
    """

    @classmethod
    def new(cls, depc, qaid_list, daid_list, cfgdict=None, tablename=None):
        parent_rowids = list(zip(qaid_list))
        request = cls.static_new(cls, depc, parent_rowids, cfgdict, tablename)
        request.qaids = safeop(np.array, qaid_list)
        request.daids = safeop(np.array, daid_list)
        # HACK
        request.config.daids = request.daids
        return request

    def get_input_hashid(request):
        # return '_'.join([request.get_query_hashid(), request.get_data_hashid()])
        return '_'.join([request.get_query_hashid()])

    def get_cfgstr(
        request, with_input=False, with_data=True, with_pipe=True, hash_pipe=False
    ):
        r"""
        Override default get_cfgstr to show reliance on data
        """
        cfgstr_list = []
        if with_input:
            cfgstr_list.append(request.get_query_hashid())
        if with_data:
            cfgstr_list.append(request.get_data_hashid())
        if with_pipe:
            if hash_pipe:
                cfgstr_list.append(request.get_pipe_hashid())
            else:
                cfgstr_list.append(request.get_pipe_cfgstr())
        cfgstr = '_'.join(cfgstr_list)
        return cfgstr

    def __nice__(request):
        dbname = (
            None
            if request.depc is None or request.depc.controller is None
            else request.depc.controller.get_dbname()
        )
        infostr_ = 'nQ=%s, nD=%s %s' % (
            len(request.qaids),
            len(request.daids),
            request.get_pipe_hashid(),
        )
        return '(%s) %s' % (dbname, infostr_)


class ClassVsClassSimilarityRequest(BaseRequest):
    pass


class AlgoResult(object):
    """ Base class for algo result objects """

    @classmethod
    def load_from_fpath(cls, fpath, verbose=ut.VERBOSE):
        state_dict = ut.load_cPkl(fpath, verbose=verbose)
        self = cls()
        self.__setstate__(state_dict)
        return self

    def save_to_fpath(cm, fpath, verbose=ut.VERBOSE):
        ut.save_cPkl(fpath, cm.__getstate__(), verbose=verbose, n=2)

    def __getstate__(self):
        state_dict = self.__dict__
        return state_dict

    def __setstate__(self, state_dict):
        self.__dict__.update(state_dict)

    def copy(self):
        cls = self.__class__
        out = cls()
        state_dict = copy.deepcopy(self.__getstate__())
        out.__setstate__(state_dict)
        return out


def safeop(op_, xs, *args, **kwargs):
    return None if xs is None else op_(xs, *args, **kwargs)


class MatchResult(AlgoResult, ut.NiceRepr):
    def __init__(
        self,
        qaid=None,
        daids=None,
        qnid=None,
        dnid_list=None,
        annot_score_list=None,
        unique_nids=None,
        name_score_list=None,
    ):
        self.qaid = qaid
        self.daid_list = safeop(np.array, daids)
        self.dnid_list = safeop(np.array, dnid_list)
        self.annot_score_list = safeop(np.array, annot_score_list)
        self.name_score_list = safeop(np.array, name_score_list)

    @property
    def num_daids(cm):
        return None if cm.daid_list is None else len(cm.daid_list)

    @property
    def daids(cm):
        return cm.daid_list

    @property
    def qaids(cm):
        return cm.qaid

    def __nice__(cm):
        return ' qaid=%s nD=%s' % (cm.qaid, cm.num_daids)
